import numpy as np
import openmdao.api as om

from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Dynamic


class AeroForces(om.ExplicitComponent):
    """Compute lift and drag from coefficients"""

    def initialize(self):
        self.options.declare("num_nodes", default=1, types=int)

    def setup(self):
        nn = self.options["num_nodes"]

        self.add_input("CL", 1.0, units="unitless", shape=nn, desc="Lift coefficient")
        self.add_input("CD", 1.0, units="unitless", shape=nn, desc="Drag coefficient")
        self.add_input(Dynamic.Mission.DYNAMIC_PRESSURE, 1.0,
                       units="psf", shape=nn, desc="Dynamic pressure")

        add_aviary_input(self, Aircraft.Wing.AREA, val=1370.3)

        self.add_output(Dynamic.Mission.LIFT, units="lbf", shape=nn, desc="Lift force")
        self.add_output(Dynamic.Mission.DRAG, units="lbf", shape=nn, desc="Drag force")

    def setup_partials(self):
        nn = self.options["num_nodes"]
        arange = np.arange(nn)
        self.declare_partials(
            Dynamic.Mission.LIFT, [
                "CL", Dynamic.Mission.DYNAMIC_PRESSURE], rows=arange, cols=arange)
        self.declare_partials(Dynamic.Mission.LIFT, [Aircraft.Wing.AREA])
        self.declare_partials(
            Dynamic.Mission.DRAG, [
                "CD", Dynamic.Mission.DYNAMIC_PRESSURE], rows=arange, cols=arange)
        self.declare_partials(Dynamic.Mission.DRAG, [Aircraft.Wing.AREA])

    def compute(self, inputs, outputs):
        CL, CD, q, wing_area = inputs.values()
        outputs[Dynamic.Mission.LIFT] = q * CL * wing_area
        outputs[Dynamic.Mission.DRAG] = q * CD * wing_area

    def compute_partials(self, inputs, J):
        CL, CD, q, wing_area = inputs.values()

        J[Dynamic.Mission.LIFT, "CL"] = q * wing_area
        J[Dynamic.Mission.LIFT, Dynamic.Mission.DYNAMIC_PRESSURE] = CL * wing_area
        J[Dynamic.Mission.LIFT, Aircraft.Wing.AREA] = q * CL

        J[Dynamic.Mission.DRAG, "CD"] = q * wing_area
        J[Dynamic.Mission.DRAG, Dynamic.Mission.DYNAMIC_PRESSURE] = CD * wing_area
        J[Dynamic.Mission.DRAG, Aircraft.Wing.AREA] = q * CD


class CLFromLift(om.ExplicitComponent):
    """Compute a CL from lift, used commonly for getting CL required"""

    def initialize(self):
        self.options.declare("num_nodes", default=1, types=int)

    def setup(self):
        nn = self.options["num_nodes"]
        self.add_input("lift_req", 1, units="lbf", shape=nn, desc="Lift force")
        self.add_input(Dynamic.Mission.DYNAMIC_PRESSURE, 1.0,
                       units="psf", shape=nn, desc="Dynamic pressure")

        add_aviary_input(self, Aircraft.Wing.AREA, val=1370.3)

        self.add_output("CL", shape=nn, units='unitless', desc="Lift coefficient")

    def setup_partials(self):
        ar = np.arange(self.options["num_nodes"])
        self.declare_partials(
            "CL", ["lift_req", Dynamic.Mission.DYNAMIC_PRESSURE], rows=ar, cols=ar)
        self.declare_partials("CL", [Aircraft.Wing.AREA])

    def compute(self, inputs, outputs):
        lift_req, q, wing_area = inputs.values()
        outputs["CL"] = lift_req / (q * wing_area)

    def compute_partials(self, inputs, J):
        lift_req, q, wing_area = inputs.values()
        J["CL", "lift_req"] = 1 / (q * wing_area)
        J["CL", Dynamic.Mission.DYNAMIC_PRESSURE] = -lift_req / (q**2 * wing_area)
        J["CL", Aircraft.Wing.AREA] = -lift_req / (q * wing_area**2)


class TimeRamp(om.ExplicitComponent):
    """Ramp up or down an output over time."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._last_index = None

    def initialize(self):
        self.options.declare("num_inputs", default=1, types=int)
        self.options.declare("num_nodes", default=1, types=int)
        self.options.declare(
            "ramp_up",
            default=True,
            types=bool,
            desc="True to ramp up from 0, False to ramp down to 0",
        )
        self.options.declare(
            "units",
            default="unitless",
            types=str,
            desc="Units of the input and output variables",
        )

    def setup(self):
        k = self.options["num_inputs"]
        nn = self.options["num_nodes"]
        units = self.options["units"]

        self.add_input("t_init", 0.0, units="s", desc="Time at which to start the ramp")
        self.add_input("t_curr", 0.0, units="s", shape=nn, desc="Current time")
        self.add_input("duration", 1.0, units="s", desc="Time duration of the ramp")
        self.add_input(
            "x", 1.0, shape=(k, nn), units=units, desc="Input values in the 'on' condition"
        )

        self.add_output("y", shape=(k, nn), units=units,
                        desc="Input values ramped over time")

        # Declare partials, taking advantage of sparsity.

        row_col = np.arange(k * nn)
        self.declare_partials(of="y", wrt="x", rows=row_col, cols=row_col)

        col = np.tile(np.arange(nn), k)
        self.declare_partials(of="y", wrt="t_curr", rows=row_col, cols=col)

        self.declare_partials(of="y", wrt=["t_init", "duration"])

    def compute(self, inputs, outputs):
        x = inputs["x"]
        y = outputs["y"]
        t_init = inputs["t_init"]
        t_curr = inputs["t_curr"]
        duration = inputs["duration"]

        t_end = t_init + duration
        idx = np.searchsorted(t_curr.real, [t_init.real, t_end.real])
        i1 = idx[0, 0]
        i2 = idx[1, 0]
        self._last_index = (i1, i2)

        fact = 1.0 / duration
        scale = (t_curr[i1:i2] - t_init) * fact

        if self.options["ramp_up"]:
            y[:, :i1] = 0
            y[:, i2:] = x[:, i2:]
            y[:, i1:i2] = scale * x[:, i1:i2]
        else:
            y[:, :i1] = x[:, :i1]
            y[:, i2:] = 0
            y[:, i1:i2] = (1 - scale) * x[:, i1:i2]

    def compute_partials(self, inputs, J):
        nn = self.options["num_nodes"]
        k = self.options["num_inputs"]

        x = inputs["x"]
        t_init = inputs["t_init"]
        t_curr = inputs["t_curr"]
        duration = inputs["duration"]

        i1, i2 = self._last_index

        fact = 1.0 / duration
        delta_t = t_curr[i1:i2] - t_init
        scale = delta_t * fact

        d_x = np.zeros(nn)
        d_tcurr = x.copy()

        if self.options["ramp_up"]:
            d_x[i2:] = 1.0
            d_x[i1:i2] = scale

            d_tcurr[:, :i1] = 0.0
            d_tcurr[:, i2:] = 0.0
            d_tcurr[:, i1:i2] *= fact

            d_duration = x * (t_curr - t_init) * (-1.0 / duration**2)
            d_duration[:, :i1] = 0.0
            d_duration[:, i2:] = 0.0

            d_tinit = x * -fact
            d_tinit[:, :i1] = 0.0
            d_tinit[:, i2:] = 0.0

        else:
            d_x[:i1] = 1.0
            d_x[i1:i2] = 1 - scale

            d_tcurr[:, :i1] = 0.0
            d_tcurr[:, i2:] = 0.0
            d_tcurr[:, i1:i2] *= -fact

            d_duration = x * (t_curr - t_init) * (1.0 / duration**2)
            d_duration[:, :i1] = 0.0
            d_duration[:, i2:] = 0.0

            d_tinit = x * fact
            d_tinit[:, :i1] = 0.0
            d_tinit[:, i2:] = 0.0

        J["y", "x"] = np.tile(d_x, k)
        J["y", "t_curr"] = d_tcurr.ravel()
        J["y", "duration"] = d_duration.ravel()
        J["y", "t_init"] = d_tinit.ravel()


class TanhRampComp(om.ExplicitComponent):
    """ Differentiable transition from one steady state condition to another using a hyperbolic tangent.
    The hyperbolic tangent activation function is:
    r_1(t) = np.tanh(t)
    This has a response centered about t=0, and a value of (-0.996, 0.996) at (-np.pi, np.pi) (thus the duration
    is roughly 2*np.pi.
    This implementation allows that response to be shifted and stretch such that the user can specify:
    1. the initial steady state value before the ramp.
    2. the final steady state value after the ramp.
    3. the starting time of the ramp (where we assume the nominal start time is -np.pi).
    4. the duration of the ramp (where we assume the nominal duration of the ramp is 2 * np.pi.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._ramps = {}

    def add_ramp(self, output_name, output_units=None, shape=(1,), initial_val=0.0,
                 final_val=1.0, t_init_val=0.0, t_duration_val=1.0):
        """
        Add a tanh ramp function with the given output name to the component.
        Parameters
        ----------
        output_name : str
            The name of the output response variable.
        output_units : str or None
            Units of the response variable.
        shape : tuple
            Shape of the variable subject to the ramp at each node.
        initial_val : float
            The initial value that the ramp asymptotically approaches backward in time.
        final_val : float
            The final value that the ramp asymptotically approaches forward in time.
        t_init_val : float
            The default value for the time at which the ramp is initiated. Note that the value asymptotically
            departs the initial value and so it nearly but not exactly the initial value at this point.
        t_duration_val : float
            The default value for the time after t_init_val at which the ramp is approximately equal to the desired
            final value. Again, the hyperbolic tangent function will never exactly equal the final value but it is
            relatively flat after this duration is expired.
        """
        self._ramps[output_name] = {'shape': shape,
                                    'units': output_units,
                                    'initial_val_name': f"{output_name}:initial_val",
                                    'initial_val': initial_val,
                                    'final_val_name': f"{output_name}:final_val",
                                    'final_val': final_val,
                                    't_init_name': f"{output_name}:t_init",
                                    't_init_val': t_init_val,
                                    't_duration_name': f"{output_name}:t_duration",
                                    't_duration_val': t_duration_val}

    def initialize(self):
        self.options.declare("num_nodes", types=int)
        self.options.declare("time_units", types=str, default='s', allow_none=True)

    def setup(self):
        nn = self.options["num_nodes"]
        ar = np.arange(nn, dtype=int)

        self.add_input("time", units=self.options["time_units"], shape=(nn,))

        for output_name, options in self._ramps.items():
            size = np.prod(options['shape'], dtype=int)
            cs = np.tile(np.arange(size, dtype=int), nn)

            self.add_output(output_name, shape=(nn,) +
                            options['shape'], units=options['units'])

            self.add_input(options['initial_val_name'], val=options['initial_val']
                           * np.ones(options['shape']), units=options['units'])
            self.add_input(options['final_val_name'], val=options['final_val']
                           * np.ones(options['shape']), units=options['units'])
            self.add_input(
                options['t_init_name'], val=options['t_init_val'], units=self.options['time_units'])
            self.add_input(
                options['t_duration_name'], val=options['t_duration_val'], units=self.options['time_units'])

            self.declare_partials(
                of=output_name, wrt=f"{output_name}:initial_val", rows=ar, cols=cs)
            self.declare_partials(
                of=output_name, wrt=f"{output_name}:final_val", rows=ar, cols=cs)

            self.declare_partials(
                of=output_name, wrt=f"{output_name}:t_init", rows=ar, cols=np.zeros(nn, dtype=int))
            self.declare_partials(of=output_name, wrt=f"{output_name}:t_duration",
                                  rows=ar, cols=np.zeros(nn, dtype=int))
            self.declare_partials(of=output_name, wrt="time", rows=ar, cols=ar)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        t = inputs["time"]

        for name, options in self._ramps.items():
            initial_val = inputs[options["initial_val_name"]]
            final_val = inputs[options["final_val_name"]]
            t_init = inputs[options["t_init_name"]]
            t_duration = inputs[options["t_duration_name"]]

            dval = final_val - initial_val
            tanh_term = np.tanh(2 * np.pi * (t - t_init) / t_duration - np.pi)
            outputs[name] = 0.5 * dval * (1+tanh_term) + initial_val

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        t = inputs["time"]

        for name, options in self._ramps.items():
            initial_val = inputs[options["initial_val_name"]]
            final_val = inputs[options["final_val_name"]]
            t_init = inputs[options["t_init_name"]]
            t_duration = inputs[options["t_duration_name"]]

            dval = final_val - initial_val
            dvald2 = 0.5 * dval

            tanh_term = np.tanh(2 * np.pi * (t - t_init) / t_duration - np.pi)
            omtanh2 = 1 - tanh_term ** 2

            dtanh_term_dt = omtanh2 * 2 * np.pi / t_duration
            dtanh_term_dtinit = -dtanh_term_dt

            dtanh_term_dtduration = omtanh2 * \
                (-2 * np.pi * (t - t_init) / t_duration ** 2)

            partials[name, "time"] = dvald2 * dtanh_term_dt
            partials[name, options["t_init_name"]] = dvald2 * dtanh_term_dtinit
            partials[name, options["t_duration_name"]] = dvald2 * dtanh_term_dtduration
            partials[name, options["initial_val_name"]] = -0.5 * (1+tanh_term) + 1.0
            partials[name, options["final_val_name"]] = 0.5 * (1+tanh_term)



"""
Test the conversion between phase info and phase builder to ensure
consistency and correctness.
"""
import unittest
from copy import deepcopy

from openmdao.utils.assert_utils import assert_near_equal

from aviary.interface.default_phase_info.two_dof import phase_info as ph_in_two_dof
from aviary.interface.default_phase_info.two_dof import phase_info_parameterization as phase_info_parameterization_two_dof
from aviary.interface.default_phase_info.height_energy import phase_info as ph_in_height_energy
from aviary.interface.default_phase_info.height_energy import phase_info_parameterization as phase_info_parameterization_height_energy
from aviary.interface.methods_for_level2 import AviaryProblem

from aviary.mission.phase_builder_base import \
    PhaseBuilderBase as PhaseBuilder, phase_info_to_builder
from aviary.variable_info.variables import Aircraft, Mission


class TestPhaseInfo(unittest.TestCase):

    def _test_phase_info_dict(self, phase_info_dict, name):
        """Helper method to test a given phase_info dict"""

        _climb_info = (name, phase_info_dict[name])

        # Removing the 'fix_duration' key from user_options for test comparison
        _climb_info[1]['user_options'].pop('fix_duration')

        # Convert phase info to a phase builder
        _phase_builder: PhaseBuilder = phase_info_to_builder(*_climb_info)

        # Convert back the phase builder to phase info
        _phase_builder_info = _phase_builder.to_phase_info()

        if _climb_info != _phase_builder_info:
            lhs_name, lhs_info = _climb_info
            rhs_name, rhs_info = _phase_builder_info

            if lhs_name != rhs_name:
                raise RuntimeError(f'name mismatch: {lhs_name} != {rhs_name}')

            lhs_keys = set(sorted(set(lhs_info.keys())))
            rhs_keys = set(sorted(set(rhs_info.keys())))

            common = lhs_keys & rhs_keys
            lhs_unique = lhs_keys - common

            if lhs_keys != rhs_keys:
                if not common:
                    raise RuntimeError(
                        f'key mismatch: no keys in common: {lhs_keys} != {rhs_keys}')
                if lhs_unique:
                    raise RuntimeError(
                        'key mismatch: the new builder is missing the following keys:' f' {lhs_unique}')

            # Loop through each key to compare values
            for key in lhs_keys:
                lhs_value = lhs_info[key]
                rhs_value = rhs_info[key]

                if lhs_value != rhs_value:
                    if key in ['user_options', 'initial_guesses']:
                        for name in lhs_value:
                            lhs_option = lhs_value[name]
                            rhs_option = rhs_value[name]

                            if lhs_option != rhs_option:
                                raise RuntimeError(
                                    f'value mismatch ({key}[{name}]):' f' {lhs_option} != {rhs_option}')
                    else:
                        raise RuntimeError(
                            f'value mismatch ({key}): {lhs_value} != {rhs_value}')

    def test_default_phase_height_energy(self):
        """Tests the roundtrip conversion for default_phase_info.height_energy"""
        from aviary.interface.default_phase_info.height_energy import phase_info
        local_phase_info = deepcopy(phase_info)
        self._test_phase_info_dict(local_phase_info, 'cruise')


class TestParameterizePhaseInfo(unittest.TestCase):

    def test_phase_info_parameterization_two_dof(self):
        phase_info = deepcopy(ph_in_two_dof)

        prob = AviaryProblem()

        csv_path = "models/small_single_aisle/small_single_aisle_GwGm.csv"

        prob.load_inputs(csv_path, phase_info)
        prob.check_and_preprocess_inputs()

        # We can set some crazy vals, since we aren't going to optimize.
        prob.aviary_inputs.set_val(Mission.Design.RANGE, 5000, 'km')
        prob.aviary_inputs.set_val(Mission.Design.CRUISE_ALTITUDE, 31000, units='ft')
        prob.aviary_inputs.set_val(Mission.Design.GROSS_MASS, 120000, 'lbm')
        prob.aviary_inputs.set_val(Mission.Design.MACH, 0.6, 'unitless')

        prob.add_pre_mission_systems()
        prob.add_phases(phase_info_parameterization=phase_info_parameterization_two_dof)
        prob.add_post_mission_systems()

        prob.link_phases()

        prob.setup()
        prob.set_initial_guesses()

        prob.run_model()

        assert_near_equal(prob.get_val("traj.desc2.timeseries.input_values:distance", units='km')[-1],
                          5000.0)
        assert_near_equal(prob.get_val("traj.climb2.timeseries.input_values:altitude", units='ft')[-1],
                          31000.0)
        assert_near_equal(prob.get_val("traj.groundroll.timeseries.input_values:mass", units='lbm')[0],
                          120000.0)
        assert_near_equal(prob.get_val("traj.cruise.rhs.mach")[0],
                          0.6)

    def test_phase_info_parameterization_height_energy(self):
        phase_info = deepcopy(ph_in_height_energy)

        prob = AviaryProblem()

        csv_path = "models/test_aircraft/aircraft_for_bench_FwFm.csv"

        prob.load_inputs(csv_path, phase_info)
        prob.check_and_preprocess_inputs()

        # We can set some crazy vals, since we aren't going to optimize.
        prob.aviary_inputs.set_val(Mission.Design.RANGE, 5000., 'km')
        prob.aviary_inputs.set_val(Mission.Design.CRUISE_ALTITUDE, 31000., units='ft')
        prob.aviary_inputs.set_val(Mission.Design.GROSS_MASS, 195000., 'lbm')
        prob.aviary_inputs.set_val(Mission.Summary.CRUISE_MACH, 0.6, 'unitless')

        prob.add_pre_mission_systems()
        prob.add_phases(
            phase_info_parameterization=phase_info_parameterization_height_energy)
        prob.add_post_mission_systems()

        prob.link_phases()

        prob.setup()
        prob.set_initial_guesses()

        prob.run_model()

        range_resid = prob.get_val(Mission.Constraints.RANGE_RESIDUAL, units='km')[-1]
        assert_near_equal(range_resid, 5000.0 - 1.e-3)
        assert_near_equal(prob.get_val("traj.cruise.timeseries.altitude", units='ft')[0],
                          31000.0)
        assert_near_equal(prob.get_val("traj.cruise.timeseries.mach")[0],
                          0.6)


# To run the tests
if __name__ == '__main__':
    unittest.main()
    # test = TestPhaseInfo()
    # test.test_default_phase_height_energy()



import numpy as np
from openmdao.utils.assert_utils import assert_near_equal

from aviary.variable_info.variables import Aircraft, Dynamic, Mission


def report_cruise(prob, remote=False, file=None):
    variables = [
        ("mass", "lbm", "mass_initial", "mass_final"),
        ("alpha", "deg", "aero_initial.alpha", "aero_final.alpha"),
        ("CL", "", "aero_initial.CL", "aero_final.CL"),
        ("CD", "", "aero_initial.CD", "aero_final.CD"),
        (Dynamic.Mission.THRUST_TOTAL, "lbf", "prop_initial.thrust_net_total",
         "prop_final.thrust_net_total"),
        (Dynamic.Mission.FUEL_FLOW_RATE_TOTAL, "lbm/s",
         "fuel_flow_initial", "fuel_flow_final"),
        ("TSFC", "1/h", "prop_initial.sfc", "prop_final.sfc"),
    ]

    print(60 * "=", file=file)
    print("{:^60}".format("cruise performance"), file=file)
    print("{:<21}   {:>15}   {:>15}".format("variable", "initial", "final"), file=file)
    print(60 * "-", file=file)
    for name, unstr, pathi, pathf in variables:
        units = None if unstr == "" else unstr
        vali = prob.get_val(f"cruise.{pathi}", units=units, get_remote=remote)[0]
        valf = prob.get_val(f"cruise.{pathf}", units=units, get_remote=remote)[0]
        print(f"{name:<10} {unstr:>10} | {vali:15.4f} | {valf:15.4f}", file=file)
    print(60 * "=", file=file)


def report_fuelburn(prob, remote=False, file=None):
    traj_phases = {
        "traj": [
            "groundroll",
            "rotation",
            "ascent",
            "accel",
            "climb1",
            "climb2",
            "cruise",
            "desc1",
            "desc2",
        ]
    }

    print(40 * "=", file=file)
    print("{:^40}".format("fuel burn in each phase"), file=file)
    print(40 * "-", file=file)
    for traj, phaselist in traj_phases.items():
        for phase in phaselist:
            if phase == 'cruise':
                vals = prob.get_val(
                    f"{traj}.{phase}.timeseries.mass",
                    units="lbm",
                    indices=[-1, 0],
                    get_remote=remote,
                )
            else:
                vals = prob.get_val(
                    f"{traj}.{phase}.states:mass",
                    units="lbm",
                    indices=[-1, 0],
                    get_remote=remote,
                )
            diff = np.diff(vals, axis=0)[0, 0]
            print(f"{phase:<12} {diff:>8.2f} lbm", file=file)
    print(40 * "=", file=file)


def report_gasp_comparison(prob, rtol=0.1, remote=False, file=None):
    # values from GASP output of large single aisle 1 at constant altitude using (I think) turbofan_23k_1
    expected_vals = [
        (
            "Design GTOW (lb)",
            175400.0,
            prob.get_val(Mission.Design.GROSS_MASS, get_remote=remote)[0],
        ),
        (
            "Actual GTOW (lb)",
            175400.0,
            prob.get_val(Mission.Summary.GROSS_MASS, get_remote=remote)[0],
        ),
        (
            "OEM (lb)",
            96348.0,
            prob.get_val(Aircraft.Design.OPERATING_MASS, get_remote=remote)[0],
        ),
        # mass at end of descent
        (
            "final mass (lb)",
            137348.0,
            prob.get_val(Mission.Landing.TOUCHDOWN_MASS, get_remote=remote)[0],
        ),
        # block fuel, includes reserves
        (
            "block fuel (lb)",
            43049.6,
            prob.get_val(Mission.Summary.TOTAL_FUEL_MASS, get_remote=remote)[0],
        ),
        (
            "landing distance (ft)",
            2980.0,
            prob.get_val('landing.'+Mission.Landing.GROUND_DISTANCE,
                         get_remote=remote)[0],
        ),
        (
            "flight distance (NM)",
            3675.0,
            prob.get_val(
                "traj.desc2.timeseries.distance", units="NM", get_remote=remote
            )[-1][0],
        ),
        # ROC @ 49 seconds to get from 456.3 ft to 500 ft minus liftoff time
        (
            "ascent duration (s)",
            49 + (500 - 465.3) / (2674.4 / 60) - 31.2,
            prob.get_val("traj.ascent.t_duration", get_remote=remote)[0],
        ),
        # AEO takeoff table
        (
            "gear retraction time (s)",
            37.3,
            prob.get_val("h_fit.t_init_gear", get_remote=remote)[0],
        ),
        (
            "flaps retraction time (s)",
            47.6,
            prob.get_val("h_fit.t_init_flaps", get_remote=remote)[0],
        ),
        # block time minus taxi time, no final descent
        (
            "flight time (hr)",
            8.295 - 10 / 60,
            prob.get_val("traj.desc2.timeseries.time", units="h", get_remote=remote)[
                -1
            ][0],
        ),
    ]

    print(80 * "=", file=file)
    print("{:^80}".format("comparison with reference GASP Large Single Aisle 1"), file=file)
    print(
        "{:<28} | {:>15} | {:>15} | {:>10}".format(
            "variable", "GASP vals", "Aviary actual", "rel. err."
        ),
        file=file,
    )
    print(80 * "-", file=file)
    results = {
        item[0]: {
            "desired": item[1],
            "actual": item[2],
            "rerr": (item[2] - item[1]) / item[1],
        }
        for item in expected_vals
    }
    for label, desired, actual in expected_vals:
        mark = ""
        rerr = (actual - desired) / desired
        if abs(rerr) > rtol:
            mark = "X"
        try:
            assert_near_equal(actual, desired, tolerance=rtol)
        except ValueError:
            mark = "X"
        print(
            f"{label:<28} | {desired:15.4f} | {actual:15.4f} | {rerr:> .2e} {mark}",
            file=file,
        )
    print(80 * "=", file=file)
    return results


def report_benchmark_comparison(
    prob, rtol=0.1, remote=False, file=None, size_engine=False, base='GASP'
):
    # values from GASP output of large single aisle 1 at constant altitude using (I think) turbofan_23k_1

    no_size = [
        175310.9512,
        175310.9512,
        96457.2320,
        137455.2320,
        42853.7192,
        2615.5252,
        3675.0000,
        17.3416,
        38.2458,
        48.2047,
        8.1171,
    ]
    yes_size = [
        175539.0746,
        175539.0746,
        96587.0999,
        137585.1013,
        42951.9734,
        2568.6457,
        3675.0000,
        21.4275,
        37.9418,
        49.8459,
        8.1145,
    ]
    # no_size_yes_pyCycle = [
    #     175394.6488,
    #     175394.6488,
    #     95820.0936,
    #     136997.9090,
    #     43394.7398,
    #     2560.4481,
    #     3675.0000,
    #     10.0131,
    #     33.2441,
    #     40.2571,
    #     9.5767,
    # ]

    FLOPS_base = [
        173370.02714369,
        0.000001,  # not exactly zero to avoid divide by zero errors
        95569.5191,
        136425.51908379,
        39944.5081,
        7732.7314,
        3378.7,
    ]
    if size_engine:
        check_val = yes_size
    elif not size_engine:
        check_val = no_size
    # elif not size_engine and pyCycle_used:
    #     check_val = no_size_yes_pyCycle
    if base == 'FLOPS':
        check_val = FLOPS_base
        distance_name = 'traj.descent.timeseries.distance'
        landing_dist_name = Mission.Landing.GROUND_DISTANCE
    else:
        distance_name = "traj.desc2.timeseries.distance"
        landing_dist_name = 'landing.'+Mission.Landing.GROUND_DISTANCE

    expected_vals = [
        (
            "Design GTOW (lb)",
            check_val[0],
            prob.get_val(Mission.Design.GROSS_MASS, get_remote=remote)[0],
        ),
        (
            "Actual GTOW (lb)",
            check_val[1],
            prob.get_val(Mission.Summary.GROSS_MASS, get_remote=remote)[0],
        ),
        (
            "OEW (lb)",
            check_val[2],
            prob.get_val(Aircraft.Design.OPERATING_MASS, get_remote=remote)[0],
        ),
        # mass at end of descent
        (
            "final mass (lb)",
            check_val[3],
            prob.get_val(Mission.Landing.TOUCHDOWN_MASS, get_remote=remote)[0],
        ),
        # block fuel, includes reserves
        (
            "block fuel (lb)",
            check_val[4],
            prob.get_val(Mission.Summary.TOTAL_FUEL_MASS, get_remote=remote)[0],
        ),
        (
            "landing distance (ft)",
            check_val[5],
            prob.get_val(landing_dist_name,
                         get_remote=remote)[0],
        ),
        (
            "flight distance (NM)",
            check_val[6],
            prob.get_val(
                distance_name, units="NM", get_remote=remote
            )[-1][0],
        ),
    ]

    if base != 'FLOPS':
        GASP_vals = [
            # ROC @ 49 seconds to get from 456.3 ft to 500 ft minus liftoff time
            (
                "ascent duration (s)",
                check_val[7],
                prob.get_val("traj.ascent.t_duration", get_remote=remote)[0],
            ),
            # AEO takeoff table
            (
                "gear retraction time (s)",
                check_val[8],
                prob.get_val("h_fit.t_init_gear", get_remote=remote)[0],
            ),
            (
                "flaps retraction time (s)",
                check_val[9],
                prob.get_val("h_fit.t_init_flaps", get_remote=remote)[0],
            ),
            # block time minus taxi time, no final descent
            (
                "flight time (hr)",
                check_val[10],
                prob.get_val("traj.desc2.timeseries.time", units="h", get_remote=remote)[
                    -1
                ][0],
            ),
        ]
        expected_vals = expected_vals + GASP_vals

    print(80 * "=", file=file)
    print(
        "{:^80}".format(
            "comparison with reference value (not a validated value, just a point to check against)"
        ),
        file=file,
    )
    print(
        "{:<28} | {:>15} | {:>15} | {:>10}".format(
            "variable", "original vals", "Aviary actual", "rel. err."
        ),
        file=file,
    )
    print(80 * "-", file=file)
    for label, desired, actual in expected_vals:
        mark = ""
        rerr = (actual - desired) / desired
        if abs(rerr) > rtol:
            mark = "X"
        try:
            assert_near_equal(actual, desired, tolerance=rtol)
        except ValueError:
            mark = "X"
        print(
            f"{label:<28} | {desired:15.4f} | {actual:15.4f} | {rerr:> .2e} {mark}",
            file=file,
        )
    print(80 * "=", file=file)



import unittest
import os

import numpy as np
import openmdao
import openmdao.api as om
import pkg_resources
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
from packaging import version

from aviary.subsystems.aerodynamics.gasp_based.table_based import (
    CruiseAero, LowSpeedAero)
from aviary.variable_info.variables import Aircraft, Dynamic


class TestCruiseAero(unittest.TestCase):

    @unittest.skipIf(version.parse(openmdao.__version__) < version.parse("3.26"),
                     "Older version of OpenMDAO does not properly skip Metamodel.")
    def test_climb(self):
        prob = om.Problem()

        fp = "subsystems/aerodynamics/gasp_based/data/large_single_aisle_1_aero_free.txt"
        prob.model = CruiseAero(num_nodes=8, aero_data=fp)

        prob.setup(force_alloc_complex=True)

        prob.set_val(
            Dynamic.Mission.MACH, [
                0.381, 0.384, 0.391, 0.399, 0.8, 0.8, 0.8, 0.8])
        prob.set_val("alpha", [5.19, 5.19, 5.19, 5.18, 3.58, 3.81, 4.05, 4.18])
        prob.set_val(
            Dynamic.Mission.ALTITUDE, [
                500, 1000, 2000, 3000, 35000, 36000, 37000, 37500])
        prob.run_model()

        cl_exp = np.array(
            [0.5968, 0.5975, 0.5974, 0.5974, 0.5566, 0.5833, 0.6113, 0.6257]
        )
        cd_exp = np.array(
            [0.0307, 0.0307, 0.0307, 0.0307, 0.0296, 0.0310, 0.0326, 0.0334]
        )

        assert_near_equal(prob["CL"], cl_exp, tolerance=0.005)
        assert_near_equal(prob["CD"], cd_exp, tolerance=0.009)

        partial_data = prob.check_partials(method="cs", out_stream=None)
        assert_check_partials(partial_data, atol=4e-7, rtol=2e-7)

    @unittest.skipIf(version.parse(openmdao.__version__) < version.parse("3.26"),
                     "Older version of OpenMDAO does not properly skip Metamodel.")
    def test_cruise(self):
        prob = om.Problem()
        fp = pkg_resources.resource_filename(
            "aviary", f"subsystems/aerodynamics/gasp_based/data/large_single_aisle_1_aero_free.txt")
        prob.model = CruiseAero(num_nodes=2, aero_data=fp)
        prob.setup(force_alloc_complex=True)

        prob.set_val(Dynamic.Mission.MACH, [0.8, 0.8])
        prob.set_val("alpha", [4.216, 3.146])
        prob.set_val(Dynamic.Mission.ALTITUDE, [37500, 37500])
        prob.run_model()

        cl_exp = np.array([0.6304, 0.5059])
        cd_exp = cl_exp / np.array([18.608, 18.425])

        assert_near_equal(prob["CL"], cl_exp, tolerance=0.005)
        assert_near_equal(prob["CD"], cd_exp, tolerance=0.005)

        partial_data = prob.check_partials(method="cs", out_stream=None)
        assert_check_partials(partial_data, atol=9e-8, rtol=2e-7)


class TestLowSpeedAero(unittest.TestCase):

    # gear retraction start time at takeoff
    t_init_gear_to = 37.3
    # flap retraction start time at takeoff
    t_init_flaps_to = 47.6
    # takeoff flap deflection (deg)
    flap_defl_to = 10

    free_data = pkg_resources.resource_filename(
        "aviary", f"subsystems/aerodynamics/gasp_based/data/large_single_aisle_1_aero_free.txt")
    flaps_data = pkg_resources.resource_filename(
        "aviary", f"subsystems/aerodynamics/gasp_based/data/large_single_aisle_1_aero_flaps.txt")
    ground_data = pkg_resources.resource_filename(
        "aviary", f"subsystems/aerodynamics/gasp_based/data/large_single_aisle_1_aero_ground.txt")

    @unittest.skipIf(version.parse(openmdao.__version__) < version.parse("3.26"),
                     "Older version of OpenMDAO does not properly skip Metamodel.")
    def test_groundroll(self):
        # takeoff with flaps applied, gear down, zero alt
        prob = om.Problem()
        prob.model = LowSpeedAero(num_nodes=4,
                                  free_aero_data=self.free_data,
                                  flaps_aero_data=self.flaps_data,
                                  ground_aero_data=self.ground_data,
                                  extrapolate=True)
        prob.setup()

        prob.set_val("t_curr", [0.0, 1.0, 2.0, 3.0])
        prob.set_val(Dynamic.Mission.ALTITUDE, 0)
        prob.set_val(Dynamic.Mission.MACH, [0.0, 0.009, 0.018, 0.026])
        prob.set_val("alpha", 0)
        # TODO set q if we want to test lift/drag forces

        prob.set_val("flap_defl", self.flap_defl_to)
        prob.set_val("t_init_gear", self.t_init_gear_to)
        prob.set_val("t_init_flaps", self.t_init_flaps_to)
        prob.run_model()

        cl_exp = 0.5597 * np.ones(4)
        cd_exp = 0.0572 * np.ones(4)

        # TODO yikes @ tolerances
        assert_near_equal(prob["CL"], cl_exp, tolerance=0.1)
        assert_near_equal(prob["CD"], cd_exp, tolerance=0.3)

        partial_data = prob.check_partials(
            method="fd", out_stream=None
        )  # fd because there is a cs in the time ramp
        assert_check_partials(partial_data, atol=3e-7, rtol=6e-5)

    @unittest.skipIf(version.parse(openmdao.__version__) < version.parse("3.26"),
                     "Older version of OpenMDAO does not properly skip Metamodel.")
    def test_takeoff(self):
        # takeoff crossing flap retraction and gear retraction points
        prob = om.Problem()
        prob.model = LowSpeedAero(num_nodes=8,
                                  free_aero_data=self.free_data,
                                  flaps_aero_data=self.flaps_data,
                                  ground_aero_data=self.ground_data,
                                  extrapolate=True)
        prob.setup()

        prob.set_val(
            "t_curr",
            [37.0, 38.0, 39.0, 40.0, 47.0, 48.0, 49.0, 50.0],
        )

        alts = [44.2, 62.7, 84.6, 109.7, 373.0, 419.4, 465.3, 507.8]
        prob.set_val(Dynamic.Mission.ALTITUDE, alts)
        prob.set_val(
            Dynamic.Mission.MACH, [
                0.257, 0.260, 0.263, 0.265, 0.276, 0.277, 0.279, 0.280])
        prob.set_val("alpha", [8.94, 8.74, 8.44, 8.24, 6.45, 6.34, 6.76, 7.59])
        # TODO set q if we want to test lift/drag forces

        prob.set_val(Aircraft.Wing.AREA, 1370)
        prob.set_val(Aircraft.Wing.SPAN, 117.8)

        prob.set_val("flap_defl", self.flap_defl_to)
        prob.set_val("t_init_gear", self.t_init_gear_to)
        prob.set_val("t_init_flaps", self.t_init_flaps_to)
        prob.run_model()

        cl_exp = np.array(
            [1.3734, 1.3489, 1.3179, 1.2979, 1.1356, 1.0645, 0.9573, 0.8876]
        )
        cd_exp = np.array(
            [0.1087, 0.1070, 0.1019, 0.0969, 0.0661, 0.0641, 0.0644, 0.0680]
        )

        # TODO yikes @ tolerances
        assert_near_equal(prob["CL"], cl_exp, tolerance=0.02)
        assert_near_equal(prob["CD"], cd_exp, tolerance=0.09)

        partial_data = prob.check_partials(
            method="fd", out_stream=None
        )  # fd because there is a cs in the time ramp
        assert_check_partials(
            partial_data, atol=0.255, rtol=5e-7
        )  # fd does very poorly with the t_curr, t_init, and duration values in the time
        # ramp because its step is so much bigger that cs. By decreasing the fd step
        # size you can see that the derivatives are right wrt these values


if __name__ == "__main__":
    unittest.main()



import warnings

import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class StrutGeom(om.ExplicitComponent):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        add_aviary_input(self, Aircraft.Wing.AREA, val=150)
        add_aviary_input(self, Aircraft.Strut.AREA_RATIO, val=.2)

        add_aviary_input(self, Aircraft.Strut.ATTACHMENT_LOCATION, val=0, units="ft")

        add_aviary_input(self, Aircraft.Fuselage.AVG_DIAMETER, val=10.0)

        self.add_output(
            "strut_y", val=30, units="ft", desc="YSTRUT: attachment location of strut"
        )
        add_aviary_output(self, Aircraft.Strut.LENGTH, val=1.0)
        add_aviary_output(self, Aircraft.Strut.AREA, val=30)
        add_aviary_output(self, Aircraft.Strut.CHORD, val=1.0)

    def setup_partials(self):

        self.declare_partials(
            "strut_y", [Aircraft.Strut.ATTACHMENT_LOCATION], val=0
        )
        self.declare_partials(
            Aircraft.Strut.LENGTH,
            [
                Aircraft.Strut.ATTACHMENT_LOCATION,
                Aircraft.Fuselage.AVG_DIAMETER,
            ],
            val=0,
        )
        self.declare_partials(
            Aircraft.Strut.CHORD,
            [
                Aircraft.Wing.AREA,
                Aircraft.Strut.AREA_RATIO,
                Aircraft.Strut.ATTACHMENT_LOCATION,
                Aircraft.Fuselage.AVG_DIAMETER,
            ],
            val=0,
        )
        self.declare_partials(
            Aircraft.Strut.AREA,
            [
                Aircraft.Wing.AREA,
                Aircraft.Strut.AREA_RATIO
            ],
            val=0,
        )

    def compute(self, inputs, outputs):

        strut_x = inputs[Aircraft.Strut.ATTACHMENT_LOCATION]
        wing_area = inputs[Aircraft.Wing.AREA]
        strut_wing_area_ratio = inputs[Aircraft.Strut.AREA_RATIO]
        cabin_width = inputs[Aircraft.Fuselage.AVG_DIAMETER]

        outputs["strut_y"] = strut_x / 2

        if outputs["strut_y"] < cabin_width / 2:
            warnings.warn(
                "Strut attachment location on wing is inside fuselage, did you mean to do that?")

        outputs[Aircraft.Strut.LENGTH] = np.sqrt(
            (outputs["strut_y"] - cabin_width / 2) ** 2 + cabin_width**2
        )
        outputs[Aircraft.Strut.AREA] = strut_area = wing_area * strut_wing_area_ratio
        outputs[Aircraft.Strut.CHORD] = 0.5 * strut_area / outputs[Aircraft.Strut.LENGTH]

    def compute_partials(self, inputs, partials):

        strut_x = inputs[Aircraft.Strut.ATTACHMENT_LOCATION]
        wing_area = inputs[Aircraft.Wing.AREA]
        strut_wing_area_ratio = inputs[Aircraft.Strut.AREA_RATIO]
        cabin_width = inputs[Aircraft.Fuselage.AVG_DIAMETER]

        dstruty_dstrutx = partials[
            "strut_y", Aircraft.Strut.ATTACHMENT_LOCATION
        ] = (1 / 2)
        strut_y = strut_x / 2
        strut_length = np.sqrt((strut_y - cabin_width / 2) ** 2 + cabin_width**2)

        dlength_dstrutx = partials[
            Aircraft.Strut.LENGTH, Aircraft.Strut.ATTACHMENT_LOCATION
        ] = (
            0.5
            * ((strut_y - cabin_width / 2) ** 2 + cabin_width**2) ** -0.5
            * 2
            * (strut_y - cabin_width / 2)
            * dstruty_dstrutx
        )
        dlength_dwidth = partials[Aircraft.Strut.LENGTH, Aircraft.Fuselage.AVG_DIAMETER] = (
            5 * cabin_width - 2 * strut_y
        ) / (
            2
            * (
                np.sqrt(
                    5 * cabin_width**2 - 4 * cabin_width * strut_y + 4 * strut_y**2
                )
            )
        )

        partials[Aircraft.Strut.CHORD, Aircraft.Strut.ATTACHMENT_LOCATION] = (
            -1 * 0.5 * wing_area * strut_wing_area_ratio / strut_length**2 * dlength_dstrutx
        )
        partials[Aircraft.Strut.CHORD, Aircraft.Fuselage.AVG_DIAMETER] = (
            -1 * 0.5 * wing_area * strut_wing_area_ratio / strut_length**2 * dlength_dwidth
        )
        partials[Aircraft.Strut.CHORD, Aircraft.Wing.AREA] = 0.5 / \
            strut_length * strut_wing_area_ratio
        partials[Aircraft.Strut.CHORD, Aircraft.Strut.AREA_RATIO] = 0.5 / \
            strut_length * wing_area

        partials[Aircraft.Strut.AREA, Aircraft.Wing.AREA] = strut_wing_area_ratio
        partials[Aircraft.Strut.AREA, Aircraft.Strut.AREA_RATIO] = wing_area



import unittest

import numpy as np
import openmdao.api as om
from dymos.models.atmosphere.atmos_1976 import USatm1976Comp
from openmdao.utils.assert_utils import (assert_check_partials,
                                         assert_near_equal)

from aviary.constants import RHO_SEA_LEVEL_METRIC
from aviary.mission.gasp_based.ode.unsteady_solved.unsteady_solved_flight_conditions import \
    UnsteadySolvedFlightConditions
from aviary.variable_info.enums import SpeedType
from aviary.variable_info.variables import Dynamic, Mission


class TestUnsteadyFlightConditions(unittest.TestCase):

    def _test_unsteady_flight_conditions(self, ground_roll=False, input_speed_type=SpeedType.TAS):
        nn = 5

        p = om.Problem()

        p.model.add_subsystem("atmos",
                              USatm1976Comp(num_nodes=nn),
                              promotes_inputs=["*", ("h", Dynamic.Mission.ALTITUDE)],
                              promotes_outputs=["*", ("sos", Dynamic.Mission.SPEED_OF_SOUND)])

        p.model.add_subsystem("fc",
                              UnsteadySolvedFlightConditions(num_nodes=nn,
                                                             ground_roll=ground_roll,
                                                             input_speed_type=input_speed_type),
                              promotes_inputs=["*"],
                              promotes_outputs=["*"])

        p.setup(force_alloc_complex=True)

        if input_speed_type is SpeedType.TAS:
            p.set_val(Dynamic.Mission.ALTITUDE, 37500, units="ft")
            p.set_val("TAS", 250, units="kn")
            p.set_val("dTAS_dr", np.zeros(nn), units="kn/km")
        elif input_speed_type is SpeedType.EAS:
            p.set_val(Dynamic.Mission.ALTITUDE, 37500, units="ft")
            p.set_val("EAS", 250, units="kn")
            p.set_val("dEAS_dr", np.zeros(nn), units="kn/km")
        else:
            p.set_val(Dynamic.Mission.ALTITUDE, 37500, units="ft")
            p.set_val(Dynamic.Mission.MACH, 0.78, units="unitless")
            p.set_val("dmach_dr", np.zeros(nn), units="unitless/km")

        p.run_model()

        mach = p.get_val(Dynamic.Mission.MACH)
        eas = p.get_val("EAS")
        tas = p.get_val("TAS", units="m/s")
        sos = p.get_val(Dynamic.Mission.SPEED_OF_SOUND, units="m/s")
        rho = p.get_val("rho", units="kg/m**3")
        rho_sl = RHO_SEA_LEVEL_METRIC
        dTAS_dt_approx = p.get_val("dTAS_dt_approx")

        assert_near_equal(mach, tas/sos)
        assert_near_equal(eas, tas * np.sqrt(rho / rho_sl))
        assert_near_equal(dTAS_dt_approx, np.zeros(nn))

        p.run_model()

        with np.printoptions(linewidth=1024):
            cpd = p.check_partials(method="cs")
        assert_check_partials(cpd)

    def test_unsteady_flight_conditions(self):

        for ground_roll in True, False:
            for in_type in [SpeedType.TAS, SpeedType.EAS, SpeedType.MACH]:
                with self.subTest(msg=f"ground_roll={ground_roll} in_type={in_type}"):
                    self._test_unsteady_flight_conditions(
                        ground_roll=ground_roll, input_speed_type=in_type)



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.aerodynamics.gasp_based.common import (CLFromLift,
                                                              TanhRampComp,
                                                              TimeRamp)


class TestTimeRamp(unittest.TestCase):
    def test_single_ramp_up(self):
        k = 1
        nn = 11
        t = np.linspace(0, 10, nn)
        x = np.ones(nn)
        y_exp = np.array([[0.0, 0.0, 0.0, 0.0, 0.25, 0.5, 0.75, 1.0, 1.0, 1.0, 1.0]])

        tr = TimeRamp(num_nodes=nn, num_inputs=k, ramp_up=True)

        prob = om.Problem()
        prob.model.add_subsystem("comp", tr, promotes=["*"])
        prob.setup(force_alloc_complex=True)

        prob.set_val("t_init", 3)
        prob.set_val("t_curr", t)
        prob.set_val("duration", 4)
        prob.set_val("x", x)

        prob.run_model()

        y_exp = np.array([[0.0, 0.0, 0.0, 0.0, 0.25, 0.5, 0.75, 1.0, 1.0, 1.0, 1.0]])
        assert_near_equal(prob["y"], y_exp)

        partial_data = prob.check_partials(method="cs", out_stream=None)
        assert_check_partials(partial_data, atol=1e-15, rtol=1e-15)

    def test_multi_ramp_down(self):
        k = 2
        nn = 11
        t = np.linspace(0, 10, nn)
        x = np.tile([[1], [10]], nn)
        y_exp = np.array(
            [
                [1.0, 1.0, 1.0, 1.0, 0.8, 0.6, 0.4, 0.2, 0.0, 0.0, 0.0],
                [10.0, 10.0, 10.0, 10.0, 8.0, 6.0, 4.0, 2.0, 0.0, 0.0, 0.0],
            ]
        )

        tr = TimeRamp(num_nodes=nn, num_inputs=k, ramp_up=False)

        prob = om.Problem()
        prob.model.add_subsystem("comp", tr, promotes=["*"])
        prob.setup(force_alloc_complex=True)

        prob.set_val("t_init", 3)
        prob.set_val("t_curr", t)
        prob.set_val("duration", 5)
        prob.set_val("x", x)

        prob.run_model()

        assert_near_equal(prob["y"], y_exp)

        partial_data = prob.check_partials(method="cs", out_stream=None)
        assert_check_partials(partial_data, atol=1e-15, rtol=1e-15)


class TestCLFromLift(unittest.TestCase):
    def test_partials(self):
        prob = om.Problem()
        prob.model.add_subsystem("comp", CLFromLift(num_nodes=2), promotes=["*"])
        prob.setup(check=False, force_alloc_complex=True)
        partial_data = prob.check_partials(method="cs", out_stream=None)
        assert_check_partials(partial_data, atol=1e-15, rtol=1e-15)


class TestTanhRampComp(unittest.TestCase):

    def test_tanh_ramp_up(self):
        p = om.Problem()

        nn = 1000

        c = TanhRampComp(time_units='s', num_nodes=nn)

        c.add_ramp('thruput', output_units='kg/s', initial_val=30,
                   final_val=40, t_init_val=25, t_duration_val=5)

        p.model.add_subsystem('tanh_ramp', c)

        p.setup(force_alloc_complex=True)

        p.set_val('tanh_ramp.time', val=np.linspace(0, 100, nn))

        p.run_model()

        cpd = p.check_partials(compact_print=True, method='cs')

        thruput = p.get_val('tanh_ramp.thruput')

        assert_near_equal(thruput[250], desired=30, tolerance=0.01)
        assert_near_equal(thruput[275], desired=35, tolerance=0.01)
        assert_near_equal(thruput[300], desired=40, tolerance=0.01)

        assert_near_equal(thruput[500:], desired=40*np.ones((500, 1)), tolerance=0.01)
        assert_near_equal(thruput[:200], desired=30*np.ones((200, 1)), tolerance=0.01)

        self.assertTrue(np.all(thruput >= 30))
        self.assertTrue(np.all(thruput <= 40))

        assert_check_partials(cpd, atol=1.0E-9, rtol=1.0E-12)

    def test_tanh_ramp_down(self):
        p = om.Problem()

        nn = 1000

        c = TanhRampComp(time_units='s', num_nodes=nn)

        c.add_ramp('thruput', output_units='kg/s', initial_val=40,
                   final_val=30, t_init_val=25, t_duration_val=5)

        p.model.add_subsystem('tanh_ramp', c)

        p.setup(force_alloc_complex=True)

        p.set_val('tanh_ramp.time', val=np.linspace(0, 100, nn))

        p.run_model()

        cpd = p.check_partials(compact_print=True, method='cs', out_stream=None)

        thruput = p.get_val('tanh_ramp.thruput')[:, 0]

        assert_near_equal(thruput[250], desired=40, tolerance=0.01)
        assert_near_equal(thruput[275], desired=35, tolerance=0.01)
        assert_near_equal(thruput[300], desired=30, tolerance=0.01)

        assert_near_equal(thruput[500:], desired=30*np.ones(500), tolerance=0.01)
        assert_near_equal(thruput[:200], desired=40*np.ones(200), tolerance=0.01)

        self.assertTrue(np.all(thruput >= 30))
        self.assertTrue(np.all(thruput <= 40))

        assert_check_partials(cpd, atol=1.0E-9, rtol=1.0E-12)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import openmdao.api as om
from numpy.polynomial import Polynomial


class PolynomialFit(om.ImplicitComponent):
    def initialize(self):
        self.options.declare("N_cp", types=int)

    def setup(self):

        # locations data up want to fit, which can
        # also be thought of as the "control points"
        # of the fit function
        self.add_input("h_cp", shape=self.options["N_cp"], units="ft")
        self.add_input("time_cp", shape=self.options["N_cp"], units="s")

        self.add_input("t_init_gear", 37.3, units="s")
        self.add_output("h_init_gear", shape=1, units="ft")

        self.add_input("t_init_flaps", 47.5, units="s")
        self.add_output("h_init_flaps", shape=1, units="ft")

        # these are the coefficients of the polynomial function you are fitting
        self.add_output("A", np.zeros(4))  # assuming a 5th order polynomial

        # analytic derivatives are left as an exercize for the Daniel
        # using CS here will give accurate partials, but will miss the sparsity pattern
        self.declare_partials("*", "*", method="cs")

        self.linear_solver = om.DirectSolver()

    def solve_nonlinear(self, inputs, outputs):
        X_cp = inputs["time_cp"]
        Y_cp = inputs["h_cp"]

        # using the numpy fitting function because its faster and more stable
        polynomial = Polynomial.fit(X_cp, Y_cp, deg=3)

        coeffs = polynomial.convert().coef

        outputs["A"][:] = coeffs

        (
            a0,
            a1,
            a2,
            a3,
        ) = coeffs

        x_gear = inputs["t_init_gear"]
        outputs["h_init_gear"] = a0 + a1 * x_gear + a2 * x_gear**2 + a3 * x_gear**3

        x_flaps = inputs["t_init_flaps"]
        outputs["h_init_flaps"] = (
            a0 + a1 * x_flaps + a2 * x_flaps**2 + a3 * x_flaps**3
        )

    def apply_nonlinear(self, inputs, outputs, residuals):

        (
            a0,
            a1,
            a2,
            a3,
        ) = outputs["A"]
        X_cp = inputs["time_cp"]
        Y_cp = inputs["h_cp"]

        Y_computed = a0 + a1 * X_cp + a2 * X_cp**2 + a3 * X_cp**3

        # note that derivatives are showing up in the apply_nonlinear method because
        # this is the formulation we use to form the residual.
        # We are minimizing the sum of the square of the error: np.sum((Y_computed-Y_cp)**2) w.r.t A
        # hence we differentiate the objective w.r.t A and set the resulting system of equations to 0
        d_error__d_Y_computed = 2 * (Y_computed - Y_cp)

        d_Y_computed__d_a0 = np.ones(self.options["N_cp"])
        d_Y_computed__d_a1 = X_cp
        d_Y_computed__d_a2 = X_cp**2
        d_Y_computed__d_a3 = X_cp**3

        residuals["A"][0] = np.sum(d_error__d_Y_computed * d_Y_computed__d_a0)
        residuals["A"][1] = np.sum(d_error__d_Y_computed * d_Y_computed__d_a1)
        residuals["A"][2] = np.sum(d_error__d_Y_computed * d_Y_computed__d_a2)
        residuals["A"][3] = np.sum(d_error__d_Y_computed * d_Y_computed__d_a3)

        x_gear = inputs["t_init_gear"]
        h_gear = a0 + a1 * x_gear + a2 * x_gear**2 + a3 * x_gear**3
        residuals["h_init_gear"] = h_gear - outputs["h_init_gear"]

        x_flaps = inputs["t_init_flaps"]
        h_flaps = a0 + a1 * x_flaps + a2 * x_flaps**2 + a3 * x_flaps**3
        residuals["h_init_flaps"] = h_flaps - outputs["h_init_flaps"]



import numpy as np
import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class WingBendingMass(om.ExplicitComponent):
    '''
    Calculates the mass of wing bending material. The methodology is
    based on the FLOPS weight equations, modified to output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Mission.Design.GROSS_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Wing.AEROELASTIC_TAILORING_FACTOR, val=0.0)
        add_aviary_input(self, Aircraft.Wing.BENDING_FACTOR, val=0.0)
        add_aviary_input(self, Aircraft.Wing.BENDING_MASS_SCALER, val=1.0)
        add_aviary_input(self, Aircraft.Wing.COMPOSITE_FRACTION, val=0.0)
        add_aviary_input(self, Aircraft.Wing.ENG_POD_INERTIA_FACTOR, val=0.0)
        add_aviary_input(self, Aircraft.Wing.LOAD_FRACTION, val=0.0)
        add_aviary_input(self, Aircraft.Wing.MISC_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Wing.MISC_MASS_SCALER, val=1.0)
        add_aviary_input(self, Aircraft.Wing.SHEAR_CONTROL_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Wing.SHEAR_CONTROL_MASS_SCALER, val=1.0)
        add_aviary_input(self, Aircraft.Wing.SPAN, val=0.0)
        add_aviary_input(self, Aircraft.Wing.SWEEP, val=0.0)
        add_aviary_input(self, Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=3.75)
        add_aviary_input(self, Aircraft.Wing.VAR_SWEEP_MASS_PENALTY, val=0.0)

        add_aviary_output(self, Aircraft.Wing.BENDING_MASS, val=0.0)

        self.A1 = 8.80
        self.A2 = 6.25

    def setup_partials(self):
        self.declare_partials("*", "*")

    def compute(self, inputs, outputs):
        bt = inputs[Aircraft.Wing.BENDING_FACTOR]
        ulf = inputs[Aircraft.Wing.ULTIMATE_LOAD_FACTOR]
        span = inputs[Aircraft.Wing.SPAN]
        comp_frac = inputs[Aircraft.Wing.COMPOSITE_FRACTION]
        faert = inputs[Aircraft.Wing.AEROELASTIC_TAILORING_FACTOR]
        varswp = inputs[Aircraft.Wing.VAR_SWEEP_MASS_PENALTY]
        pctl = inputs[Aircraft.Wing.LOAD_FRACTION]
        sweep = inputs[Aircraft.Wing.SWEEP]
        gross_weight = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        CAYE = inputs[Aircraft.Wing.ENG_POD_INERTIA_FACTOR]
        scaler = inputs[Aircraft.Wing.BENDING_MASS_SCALER]

        num_fuse = self.options['aviary_options'].get_val(
            Aircraft.Fuselage.NUM_FUSELAGES)

        # Note: Calculation requires weights prior to being scaled, so we need to divide
        # by the scale factor.
        W2 = inputs[Aircraft.Wing.SHEAR_CONTROL_MASS] / \
            inputs[Aircraft.Wing.SHEAR_CONTROL_MASS_SCALER] * GRAV_ENGLISH_LBM
        W3 = inputs[Aircraft.Wing.MISC_MASS] / \
            inputs[Aircraft.Wing.MISC_MASS_SCALER] * GRAV_ENGLISH_LBM

        vfact = 1.0 + varswp * (0.96 / np.cos(np.pi / 180. * sweep) - 1.0)
        cayf = 0.5 if num_fuse > 1 else 1.0

        W1NIR = self.A1 * bt * (1.0 + (self.A2 / span)**0.5) * ulf * span * \
            (1.0 - 0.4 * comp_frac) * (1.0 - 0.1 * faert) * cayf * vfact * pctl * 1.0e-6

        outputs[Aircraft.Wing.BENDING_MASS] = (
            (gross_weight * CAYE * W1NIR + W2 + W3) / (1.0 + W1NIR) - W2 - W3) \
            * scaler / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        bt = inputs[Aircraft.Wing.BENDING_FACTOR]
        ulf = inputs[Aircraft.Wing.ULTIMATE_LOAD_FACTOR]
        span = inputs[Aircraft.Wing.SPAN]
        comp_frac = inputs[Aircraft.Wing.COMPOSITE_FRACTION]
        faert = inputs[Aircraft.Wing.AEROELASTIC_TAILORING_FACTOR]
        varswp = inputs[Aircraft.Wing.VAR_SWEEP_MASS_PENALTY]
        pctl = inputs[Aircraft.Wing.LOAD_FRACTION]
        sweep = inputs[Aircraft.Wing.SWEEP]
        gross_weight = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        CAYE = inputs[Aircraft.Wing.ENG_POD_INERTIA_FACTOR]
        W2 = inputs[Aircraft.Wing.SHEAR_CONTROL_MASS] * GRAV_ENGLISH_LBM
        W3 = inputs[Aircraft.Wing.MISC_MASS] * GRAV_ENGLISH_LBM
        W2scale = inputs[Aircraft.Wing.SHEAR_CONTROL_MASS_SCALER]
        W3scale = inputs[Aircraft.Wing.MISC_MASS_SCALER]
        scaler = inputs[Aircraft.Wing.BENDING_MASS_SCALER]

        num_fuse = self.options['aviary_options'].get_val(
            Aircraft.Fuselage.NUM_FUSELAGES)

        deg2rad = np.pi / 180.
        term = 0.96 / np.cos(deg2rad * sweep)
        vfact = 1.0 + varswp * (term - 1.0)
        dvfact_varswp = (term - 1.0)
        dvfact_sweep = varswp * deg2rad * term * np.tan(deg2rad * sweep)
        cayf = 0.5 if num_fuse > 1 else 1.0

        fact0 = (self.A2 / span)**0.5
        fact1 = 1.0 + fact0
        fact2 = 1.0 - 0.4 * comp_frac
        fact3 = 1.0 - 0.1 * faert
        W1NIR = (
            self.A1 * bt * fact1 * ulf * span * fact2 * fact3 * cayf * vfact * pctl
            * 1.0e-6)
        dW1NIR_bt = \
            self.A1 * fact1 * ulf * span * fact2 * fact3 * cayf * vfact * pctl * 1.0e-6
        dW1NIR_ulf = \
            self.A1 * bt * fact1 * span * fact2 * fact3 * cayf * vfact * pctl * 1.0e-6
        dW1NIR_pctl = \
            self.A1 * bt * fact1 * ulf * span * fact2 * fact3 * cayf * vfact * 1.0e-6
        dW1NIR_cayf = \
            self.A1 * bt * fact1 * ulf * span * fact2 * fact3 * vfact * pctl * 1.0e-6
        dW1NIR_compfrac = -self.A1 * bt * fact1 * ulf * \
            span * 0.4 * fact3 * cayf * vfact * pctl * 1.0e-6
        dW1NIR_faert = -self.A1 * bt * fact1 * ulf * \
            span * fact2 * 0.1 * cayf * vfact * pctl * 1.0e-6
        dW1NIR_varswp = self.A1 * bt * fact1 * ulf * span * \
            fact2 * fact3 * cayf * dvfact_varswp * pctl * 1.0e-6
        dW1NIR_sweep = self.A1 * bt * fact1 * ulf * span * \
            fact2 * fact3 * cayf * dvfact_sweep * pctl * 1.0e-6

        dfact1_span = -0.5 * self.A2 / (fact0 * span * span)
        dW1NIR_span = self.A1 * bt * \
            (dfact1_span * span + fact1) * ulf * \
            fact2 * fact3 * cayf * vfact * pctl * 1.0e-6

        fact1 = gross_weight * CAYE * W1NIR + W2/W2scale + W3/W3scale
        fact2 = 1.0 / (1.0 + W1NIR)
        dbend_w1nir = scaler * (gross_weight * CAYE * fact2 - fact1 * fact2**2)

        J[Aircraft.Wing.BENDING_MASS, Mission.Design.GROSS_MASS] = \
            CAYE * W1NIR * fact2 * scaler

        J[Aircraft.Wing.BENDING_MASS, Aircraft.Wing.ENG_POD_INERTIA_FACTOR] = \
            gross_weight * W1NIR * fact2 * scaler / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.BENDING_MASS, Aircraft.Wing.SHEAR_CONTROL_MASS] = \
            (fact2 - 1.0) * scaler / W2scale

        J[Aircraft.Wing.BENDING_MASS, Aircraft.Wing.SHEAR_CONTROL_MASS_SCALER] = \
            -(fact2 - 1.0) * scaler * W2 / W2scale ** 2 / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.BENDING_MASS, Aircraft.Wing.MISC_MASS] = \
            (fact2 - 1.0) * scaler / W3scale

        J[Aircraft.Wing.BENDING_MASS, Aircraft.Wing.MISC_MASS_SCALER] = \
            -(fact2 - 1.0) * scaler * W3 / W3scale ** 2 / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.BENDING_MASS, Aircraft.Wing.BENDING_MASS_SCALER] = \
            (fact1 * fact2 - W2/W2scale - W3/W3scale) / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.BENDING_MASS, Aircraft.Wing.BENDING_FACTOR] = \
            dbend_w1nir * dW1NIR_bt / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.BENDING_MASS, Aircraft.Wing.ULTIMATE_LOAD_FACTOR] = \
            dbend_w1nir * dW1NIR_ulf / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.BENDING_MASS, Aircraft.Wing.LOAD_FRACTION] = \
            dbend_w1nir * dW1NIR_pctl / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.BENDING_MASS, Aircraft.Wing.COMPOSITE_FRACTION] = \
            dbend_w1nir * dW1NIR_compfrac / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.BENDING_MASS, Aircraft.Wing.AEROELASTIC_TAILORING_FACTOR] = \
            dbend_w1nir * dW1NIR_faert / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.BENDING_MASS, Aircraft.Wing.VAR_SWEEP_MASS_PENALTY] = \
            dbend_w1nir * dW1NIR_varswp / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.BENDING_MASS, Aircraft.Wing.SWEEP] = \
            dbend_w1nir * dW1NIR_sweep / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.BENDING_MASS, Aircraft.Wing.SPAN] = \
            dbend_w1nir * dW1NIR_span / GRAV_ENGLISH_LBM


class WingShearControlMass(om.ExplicitComponent):
    '''
    Calculates the mass of wing shear control material. The methodology is
    based on the FLOPS weight equations, modified to output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

        self.options.declare(
            'aircraft_type',
            default='Transport',
            values=['Transport', 'HWB', 'GA'],
            desc='Aircfaft type: Tranpsport, HWB, or GA')

    def setup(self):
        add_aviary_input(self, Aircraft.Wing.COMPOSITE_FRACTION, val=0.0)

        add_aviary_input(self, Aircraft.Wing.CONTROL_SURFACE_AREA, val=0.0)

        add_aviary_input(self, Mission.Design.GROSS_MASS, val=0.0)

        add_aviary_input(self, Aircraft.Wing.SHEAR_CONTROL_MASS_SCALER, val=1.0)

        add_aviary_output(self, Aircraft.Wing.SHEAR_CONTROL_MASS, val=0.0)

        if (
            (self.options['aircraft_type'] == 'Transport')
            or (self.options['aircraft_type'] == 'HWB')
        ):
            self.A3 = 0.68
            self.A4 = 0.34
            self.A5 = 0.60
        elif self.options['aircraft_type'] == 'GA':
            self.A3 = 0.25
            self.A4 = 0.50
            self.A5 = 0.50

    def setup_partials(self):
        self.declare_partials("*", "*")

    def compute(self, inputs, outputs):
        comp_frac = inputs[Aircraft.Wing.COMPOSITE_FRACTION]
        gross_weight = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        ctrl_area = inputs[Aircraft.Wing.CONTROL_SURFACE_AREA]
        scaler = inputs[Aircraft.Wing.SHEAR_CONTROL_MASS_SCALER]

        outputs[Aircraft.Wing.SHEAR_CONTROL_MASS] = \
            self.A3 * (1.0 - 0.17 * comp_frac) * ctrl_area ** self.A4 * \
            gross_weight ** self.A5 * scaler / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        comp_frac = inputs[Aircraft.Wing.COMPOSITE_FRACTION]
        gross_weight = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        ctrl_area = inputs[Aircraft.Wing.CONTROL_SURFACE_AREA]
        scaler = inputs[Aircraft.Wing.SHEAR_CONTROL_MASS_SCALER]

        comp_frac_term = self.A3 * (1.0 - 0.17 * comp_frac)

        J[Aircraft.Wing.SHEAR_CONTROL_MASS, Aircraft.Wing.COMPOSITE_FRACTION] = \
            -0.17 * self.A3 * ctrl_area ** self.A4 * \
            gross_weight ** self.A5 * scaler / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.SHEAR_CONTROL_MASS, Aircraft.Wing.CONTROL_SURFACE_AREA] = \
            comp_frac_term * self.A4 * \
            ctrl_area ** (self.A4-1.0) * gross_weight ** self.A5 * \
            scaler / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.SHEAR_CONTROL_MASS, Mission.Design.GROSS_MASS] = \
            comp_frac_term * ctrl_area ** self.A4 * self.A5 * \
            gross_weight ** (self.A5-1.0) * scaler

        J[
            Aircraft.Wing.SHEAR_CONTROL_MASS,
            Aircraft.Wing.SHEAR_CONTROL_MASS_SCALER] = \
            self.A3 * (1.0 - 0.17 * comp_frac) * \
            ctrl_area ** self.A4 * gross_weight ** self.A5 / GRAV_ENGLISH_LBM


class WingMiscMass(om.ExplicitComponent):
    '''
    Calculates the mass of wing miscellaneous material. The methodology is
    based on the FLOPS weight equations, modified to output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

        self.options.declare(
            'aircraft_type',
            default='Transport',
            values=['Transport', 'HWB', 'GA'],
            desc='Aircfaft type: Tranpsport, HWB, or GA')

    def setup(self):
        add_aviary_input(self, Aircraft.Wing.COMPOSITE_FRACTION, val=0.0)

        add_aviary_input(self, Aircraft.Wing.AREA, val=0.0)

        add_aviary_input(self, Aircraft.Wing.MISC_MASS_SCALER, val=1.0)

        add_aviary_output(self, Aircraft.Wing.MISC_MASS, val=0.0)

        if (
            (self.options['aircraft_type'] == 'Transport')
            or (self.options['aircraft_type'] == 'HWB')
        ):
            self.A6 = 0.035
            self.A7 = 1.50
        elif self.options['aircraft_type'] == 'GA':
            self.A6 = 0.16
            self.A7 = 1.2

    def setup_partials(self):
        self.declare_partials("*", "*")

    def compute(self, inputs, outputs):
        comp_frac = inputs[Aircraft.Wing.COMPOSITE_FRACTION]
        area = inputs[Aircraft.Wing.AREA]
        scaler = inputs[Aircraft.Wing.MISC_MASS_SCALER]

        outputs[Aircraft.Wing.MISC_MASS] = self.A6 * \
            (1.0 - 0.3 * comp_frac) * area ** self.A7 * \
            scaler / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        comp_frac = inputs[Aircraft.Wing.COMPOSITE_FRACTION]
        area = inputs[Aircraft.Wing.AREA]
        scaler = inputs[Aircraft.Wing.MISC_MASS_SCALER]

        J[Aircraft.Wing.MISC_MASS, Aircraft.Wing.COMPOSITE_FRACTION] = - \
            0.3 * self.A6 * area ** self.A7 * scaler / GRAV_ENGLISH_LBM
        J[Aircraft.Wing.MISC_MASS, Aircraft.Wing.AREA] = self.A6 * \
            (1.0 - 0.3 * comp_frac) * self.A7 * \
            area ** (self.A7-1) * scaler / GRAV_ENGLISH_LBM
        J[Aircraft.Wing.MISC_MASS, Aircraft.Wing.MISC_MASS_SCALER] = self.A6 * \
            (1.0 - 0.3 * comp_frac) * area ** self.A7 / GRAV_ENGLISH_LBM


class WingTotalMass(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Wing.BENDING_MASS, val=0.0)

        add_aviary_input(self, Aircraft.Wing.SHEAR_CONTROL_MASS, val=0.0)

        add_aviary_input(self, Aircraft.Wing.MISC_MASS, val=0.0)

        add_aviary_input(self, Aircraft.Wing.BWB_AFTBODY_MASS, val=0.0)

        add_aviary_input(self, Aircraft.Wing.MASS_SCALER, val=1.0)

        add_aviary_output(self, Aircraft.Wing.MASS, val=0)

    def setup_partials(self):
        self.declare_partials("*", "*")

    def compute(self, inputs, outputs):
        m1 = inputs[Aircraft.Wing.BENDING_MASS]
        m2 = inputs[Aircraft.Wing.SHEAR_CONTROL_MASS]
        m3 = inputs[Aircraft.Wing.MISC_MASS]
        m4 = inputs[Aircraft.Wing.BWB_AFTBODY_MASS]
        m_scalar = inputs[Aircraft.Wing.MASS_SCALER]

        outputs[Aircraft.Wing.MASS] = (m1 + m2 + m3 + m4) * m_scalar

    def compute_partials(self, inputs, J):
        m1 = inputs[Aircraft.Wing.BENDING_MASS]
        m2 = inputs[Aircraft.Wing.SHEAR_CONTROL_MASS]
        m3 = inputs[Aircraft.Wing.MISC_MASS]
        m4 = inputs[Aircraft.Wing.BWB_AFTBODY_MASS]
        m_scaler = inputs[Aircraft.Wing.MASS_SCALER]

        J[Aircraft.Wing.MASS, Aircraft.Wing.BENDING_MASS] = m_scaler
        J[Aircraft.Wing.MASS, Aircraft.Wing.SHEAR_CONTROL_MASS] = m_scaler
        J[Aircraft.Wing.MASS, Aircraft.Wing.MISC_MASS] = m_scaler
        J[Aircraft.Wing.MASS, Aircraft.Wing.BWB_AFTBODY_MASS] = m_scaler
        J[Aircraft.Wing.MASS, Aircraft.Wing.MASS_SCALER] = m1 + m2 + m3 + m4



import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class EngineScaling(om.ExplicitComponent):
    '''
    Scales an engine's thrust, fuel flow rate, nox rate, exit area, and electric power
    based on provided scaling factors.

    Can be vectorized for all unique engine present on aircraft. Each index represents a
    single instance of an engine model.
    '''

    def initialize(self):
        self.options.declare('num_nodes', types=int)

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        nn = self.options['num_nodes']
        options: AviaryValues = self.options['aviary_options']

        # count = len(options.get_val('engine_models'))  # number of unique engine models

        # add_aviary_input(self, Aircraft.Engine.SCALE_FACTOR, val=np.ones(count))
        add_aviary_input(self, Aircraft.Engine.SCALE_FACTOR, val=1.0)

        self.add_input(Dynamic.Mission.MACH, val=np.zeros(nn),
                       desc='current Mach number', units='unitless')

        # to vectorize, shape is (nn, count)
        self.add_input('thrust_net_unscaled', val=np.zeros(nn), units='lbf',
                       desc='Current net thrust produced (unscaled)')

        self.add_input('thrust_net_max_unscaled', val=np.zeros(nn),
                       units='lbf', desc='Current maximum possible net thrust (unscaled)')

        self.add_input('fuel_flow_rate_unscaled', val=np.zeros(nn),
                       units='lbm/h', desc='Current fuel flow rate (unscaled)')

        self.add_input('electric_power_unscaled', val=np.zeros(nn),
                       units='kW', desc='Current electric power consumption (unscaled)')

        self.add_input('nox_rate_unscaled', val=np.zeros(nn), units='lbm/h',
                       desc='Current NOx emission rate (unscaled)')

        # self.add_input('exit_area_unscaled', val=np.zeros(nn), units='ft**2',
        #                desc='Current engine nozzle exit area')

        self.add_output(Dynamic.Mission.THRUST, val=np.zeros(nn), units='lbf',
                        desc='Current net thrust produced')

        self.add_output(Dynamic.Mission.THRUST_MAX, val=np.zeros(nn),
                        units='lbf', desc='Current maximum possible net thrust')

        self.add_output(Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE, val=np.zeros(nn),
                        units='lbm/h', desc='Current fuel flow rate (negative)')

        self.add_output(Dynamic.Mission.ELECTRIC_POWER, val=np.zeros(nn),
                        units='kW', desc='Current electric power consumption')

        self.add_output(Dynamic.Mission.NOX_RATE, val=np.zeros(nn),
                        units='lbm/h', desc='Current NOx emission rate')

        # self.add_output(Dynamic.Mission.EXIT_AREA, val=np.zeros(nn),
        #                 units='ft**2', desc='Current engine nozzle exit area')

    def compute(self, inputs, outputs):
        nn = self.options['num_nodes']
        options: AviaryValues = self.options['aviary_options']
        # count = len(options.get_val('engine_models'))  # number of unique engine models
        scale_performance = options.get_val(Aircraft.Engine.SCALE_PERFORMANCE)

        subsonic_fuel_factor = options.get_val(Aircraft.Engine.SUBSONIC_FUEL_FLOW_SCALER)
        supersonic_fuel_factor = options.get_val(
            Aircraft.Engine.SUPERSONIC_FUEL_FLOW_SCALER)
        constant_fuel_term = options.get_val(
            Aircraft.Engine.FUEL_FLOW_SCALER_CONSTANT_TERM)
        linear_fuel_term = options.get_val(Aircraft.Engine.FUEL_FLOW_SCALER_LINEAR_TERM)
        constant_fuel_flow = options.get_val(
            Aircraft.Engine.CONSTANT_FUEL_CONSUMPTION, units='lbm/h')
        mission_fuel_scaler = options.get_val(Mission.Summary.FUEL_FLOW_SCALER)

        # thrust-based engine scaling factor
        engine_scale_factor = inputs[Aircraft.Engine.SCALE_FACTOR]

        mach_number = inputs[Dynamic.Mission.MACH]

        unscaled_net_thrust = inputs['thrust_net_unscaled']
        unscaled_max_thrust = inputs['thrust_net_max_unscaled']
        unscaled_fuel_flow_rate = inputs['fuel_flow_rate_unscaled']
        unscaled_electric_power = inputs['electric_power_unscaled']
        unscaled_nox_rate = inputs['nox_rate_unscaled']
        # unscaled_exit_area = inputs['exit_area_unscaled']

        scale_factor = 1
        fuel_flow_scale_factor = np.ones(nn, dtype=engine_scale_factor.dtype)
        # scale_idx = np.where(scale_performance)

        # if len(scale_idx[0]) > 0:
        if scale_performance:
            # special factor to scale fuel flow based on thrust when scaling is permitted
            # NOTE mission-specific fuel flow scaling factor is overwritten by
            #      scale_performance = False

            # Calculate fuel flow rate scaling factor using FLOPS-derived equation
            fuel_flow_equation_scaling = (
                1 + constant_fuel_term + linear_fuel_term * (1 - engine_scale_factor))

            # use dtype to make complex safe
            fuel_flow_mach_scaling = np.ones(
                nn, dtype=engine_scale_factor.dtype) * subsonic_fuel_factor
            supersonic_idx = np.where(mach_number >= 1.0)
            fuel_flow_mach_scaling[supersonic_idx] = supersonic_fuel_factor

            # fuel_flow_scale_factor[:, scale_idx[0]] = engine_scale_factor[scale_idx]\
            #     * fuel_flow_mach_scaling[:, scale_idx[0]]\
            #     * fuel_flow_equation_scaling[scale_idx]\
            #     * mission_fuel_scaler
            fuel_flow_scale_factor = engine_scale_factor * fuel_flow_mach_scaling\
                * fuel_flow_equation_scaling * mission_fuel_scaler

            scale_factor = engine_scale_factor

        # scale factor only applies if engine performance is scaled - default to 1 otherwise
        # scale_factor = np.ones(count, dtype=engine_scale_factor.dtype)
        # scale_factor[scale_idx] = engine_scale_factor[scale_idx]

        outputs[Dynamic.Mission.THRUST] = unscaled_net_thrust * scale_factor
        outputs[Dynamic.Mission.THRUST_MAX] = unscaled_max_thrust * scale_factor
        # user-specified constant_fuel_flow value is currently not scaled with engine
        outputs[Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE] = - \
            (unscaled_fuel_flow_rate * fuel_flow_scale_factor) - constant_fuel_flow

        # TODO determine proper scaling for electric power, nox, and exit area
        outputs[Dynamic.Mission.ELECTRIC_POWER] = unscaled_electric_power * scale_factor
        outputs[Dynamic.Mission.NOX_RATE] = unscaled_nox_rate * scale_factor
        # outputs[Dynamic.Mission.EXIT_AREA] = unscaled_exit_area * scale_factor

    def setup_partials(self):
        nn = self.options['num_nodes']
        # options = self.options['aviary_options']
        # count = len(options.get_val('engine_models'))  # number of unique engine models

        # matrix derivatives have known sparsity pattern - specified here
        # r = np.arange(nn * count, dtype=int)
        # c = np.tile(np.arange(count, dtype=int), (nn))
        r = np.arange(nn)
        c = np.tile(0, nn)

        self.declare_partials(
            Dynamic.Mission.THRUST,
            Aircraft.Engine.SCALE_FACTOR,
            rows=r, cols=c,
            val=1.0)
        self.declare_partials(
            Dynamic.Mission.THRUST,
            'thrust_net_unscaled',
            rows=r, cols=r,
            val=1.0)

        self.declare_partials(
            Dynamic.Mission.THRUST_MAX,
            Aircraft.Engine.SCALE_FACTOR,
            rows=r, cols=c,
            val=1.0)
        self.declare_partials(
            Dynamic.Mission.THRUST_MAX,
            'thrust_net_max_unscaled',
            rows=r, cols=r,
            val=1.0)

        self.declare_partials(
            Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE,
            Aircraft.Engine.SCALE_FACTOR,
            rows=r, cols=c,
            val=1.0)
        self.declare_partials(
            Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE,
            'fuel_flow_rate_unscaled',
            rows=r, cols=r,
            val=1.0)

        self.declare_partials(
            Dynamic.Mission.ELECTRIC_POWER,
            Aircraft.Engine.SCALE_FACTOR,
            rows=r, cols=c,
            val=1.0)
        self.declare_partials(
            Dynamic.Mission.ELECTRIC_POWER,
            'electric_power_unscaled',
            rows=r, cols=r,
            val=1.0)

        self.declare_partials(
            Dynamic.Mission.NOX_RATE,
            Aircraft.Engine.SCALE_FACTOR,
            rows=r, cols=c,
            val=1.0)
        self.declare_partials(
            Dynamic.Mission.NOX_RATE,
            'nox_rate_unscaled',
            rows=r, cols=r,
            val=1.0)

        # self.declare_partials(
        #     Dynamic.Mission.EXIT_AREA,
        #     Aircraft.Engine.SCALE_FACTOR,
        #     rows=r, cols=c,
        #     val=1.0)
        # self.declare_partials(
        #     Dynamic.Mission.EXIT_AREA,
        #     'exit_area_unscaled',
        #     rows=r, cols=r,
        #     val=1.0)

        # self.declare_partials(
        #     Dynamic.Mission.EXIT_AREA,
        #     Aircraft.Engine.SCALE_FACTOR,
        #     rows=r, cols=c,
        #     val=1.0)
        # self.declare_partials(
        #     Dynamic.Mission.EXIT_AREA,
        #     'exit_area_unscaled',
        #     rows=r, cols=r,
        #     val=1.0)

    def compute_partials(self, inputs, J):
        nn = self.options['num_nodes']
        options: AviaryValues = self.options['aviary_options']
        # count = len(options.get_val('engine_models'))  # number of unique engine models
        scale_performance = options.get_val(Aircraft.Engine.SCALE_PERFORMANCE)

        subsonic_fuel_factor = options.get_val(Aircraft.Engine.SUBSONIC_FUEL_FLOW_SCALER)
        supersonic_fuel_factor = options.get_val(
            Aircraft.Engine.SUPERSONIC_FUEL_FLOW_SCALER)
        constant_fuel_term = options.get_val(
            Aircraft.Engine.FUEL_FLOW_SCALER_CONSTANT_TERM, units='unitless')
        linear_fuel_term = options.get_val(Aircraft.Engine.FUEL_FLOW_SCALER_LINEAR_TERM)
        constant_fuel_flow = options.get_val(
            Aircraft.Engine.CONSTANT_FUEL_CONSUMPTION, units='lbm/h')
        mission_fuel_scaler = options.get_val(Mission.Summary.FUEL_FLOW_SCALER)

        mach_number = inputs[Dynamic.Mission.MACH]
        unscaled_net_thrust = inputs['thrust_net_unscaled']
        unscaled_max_thrust = inputs['thrust_net_max_unscaled']
        unscaled_fuel_flow_rate = inputs['fuel_flow_rate_unscaled']
        unscaled_electric_power = inputs['electric_power_unscaled']
        unscaled_nox_rate = inputs['nox_rate_unscaled']
        # unscaled_exit_area = inputs['exit_area_unscaled']

        engine_scale_factor = inputs[Aircraft.Engine.SCALE_FACTOR]

        # determine which mach-based fuel flow scaler is applied at each node
        fuel_flow_mach_scaling = np.ones(nn) * subsonic_fuel_factor
        fuel_flow_mach_scaling[mach_number >= 1.0] = supersonic_fuel_factor

        fuel_flow_deriv = np.ones(nn, dtype=engine_scale_factor.dtype)
        fuel_flow_scale_deriv = np.zeros(nn, dtype=engine_scale_factor.dtype)

        # scale factor only applies if engine performance is scaled - default to 1 otherwise
        # scale_factor = np.ones(count, dtype=engine_scale_factor.dtype)
        # scale_factor[scale_idx] = engine_scale_factor[scale_idx]
        scale_factor = 1

        # deriv_factor = np.zeros(nn)
        # deriv_factor[:, scale_idx[0]] = 1.0
        deriv_factor = 0

        # scale_idx = np.where(scale_performance)
        # if len(scale_idx[0]) > 0:
        if scale_performance:
            # Calculate fuel flow rate scaling factor using FLOPS-derived equation
            fuel_flow_equation_scaling = (
                1 + constant_fuel_term +
                linear_fuel_term * (1 - engine_scale_factor)
            )

            fuel_flow_deriv = -engine_scale_factor * fuel_flow_mach_scaling\
                * fuel_flow_equation_scaling * mission_fuel_scaler

            # fuel_flow_scale_deriv = fuel_flow_mach_scaling[:, scale_idx[0]]\
            #     * mission_fuel_scaler\
            #     * (unscaled_fuel_flow_rate[:, scale_idx[0]]
            #        + constant_fuel_flow[scale_idx]
            #        )\
            #     * (1 + linear_fuel_term[scale_idx]
            #        + constant_fuel_term[scale_idx]
            #        - (2 * linear_fuel_term[scale_idx]
            #           * engine_scale_factor[scale_idx]
            #           )
            #        )

            fuel_flow_scale_deriv = -fuel_flow_mach_scaling * mission_fuel_scaler\
                * unscaled_fuel_flow_rate \
                * (1 + linear_fuel_term + constant_fuel_term - (2 * linear_fuel_term
                                                                * engine_scale_factor))

            scale_factor = engine_scale_factor
            deriv_factor = 1.0

        # J[Dynamic.THRUST, 'thrust_net_unscaled'] = np.tile(
        #     scale_factor, nn)
        # J[Dynamic.THRUST, Aircraft.Engine.SCALE_FACTOR] = np.ravel(
        #     unscaled_net_thrust * deriv_factor)

        # J[Dynamic.THRUST_MAX, 'thrust_net_max_unscaled'] = np.tile(
        #     scale_factor, nn)
        # J[Dynamic.THRUST_MAX, Aircraft.Engine.SCALE_FACTOR] = np.ravel(
        #     unscaled_max_thrust * deriv_factor)

        # J[Dynamic.FUEL_FLOW_RATE_NEGATIVE, 'fuel_flow_rate_unscaled'] = -np.ravel(
        #     fuel_flow_deriv)
        # J[Dynamic.FUEL_FLOW_RATE_NEGATIVE, Aircraft.Engine.SCALE_FACTOR] = -np.ravel(
        #     fuel_flow_scale_deriv)

        # J[Dynamic.ELECTRIC_POWER, 'electric_power_unscaled'] = np.tile(
        #     scale_factor, nn)
        # J[Dynamic.ELECTRIC_POWER, Aircraft.Engine.SCALE_FACTOR] = np.ravel(
        #     unscaled_electric_power * deriv_factor)

        # J[Dynamic.NOX_RATE, 'nox_rate_unscaled'] = np.tile(
        #     scale_factor, nn)
        # J[Dynamic.NOX_RATE, Aircraft.Engine.SCALE_FACTOR] = np.ravel(
        #     unscaled_nox_rate * deriv_factor)

        # J[Dynamic.EXIT_AREA, 'exit_area_unscaled'] = np.tile(
        #     scale_factor, nn)
        # J[Dynamic.EXIT_AREA, Aircraft.Engine.SCALE_FACTOR] = np.ravel(
        #     unscaled_exit_area * deriv_factor)

        J[Dynamic.Mission.THRUST, 'thrust_net_unscaled'] = scale_factor
        J[Dynamic.Mission.THRUST, Aircraft.Engine.SCALE_FACTOR] = unscaled_net_thrust * deriv_factor

        J[Dynamic.Mission.THRUST_MAX, 'thrust_net_max_unscaled'] = scale_factor
        J[Dynamic.Mission.THRUST_MAX,
            Aircraft.Engine.SCALE_FACTOR] = unscaled_max_thrust * deriv_factor

        J[Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE,
            'fuel_flow_rate_unscaled'] = fuel_flow_deriv
        J[Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE,
            Aircraft.Engine.SCALE_FACTOR] = fuel_flow_scale_deriv

        J[Dynamic.Mission.ELECTRIC_POWER, 'electric_power_unscaled'] = scale_factor
        J[Dynamic.Mission.ELECTRIC_POWER,
            Aircraft.Engine.SCALE_FACTOR] = unscaled_electric_power * deriv_factor

        J[Dynamic.Mission.NOX_RATE, 'nox_rate_unscaled'] = scale_factor
        J[Dynamic.Mission.NOX_RATE, Aircraft.Engine.SCALE_FACTOR] = unscaled_nox_rate * deriv_factor

        # J[Dynamic.Mission.EXIT_AREA, 'exit_area_unscaled'] = scale_factor
        # J[Dynamic.Mission.EXIT_AREA, Aircraft.Engine.SCALE_FACTOR] = unscaled_exit_area * deriv_factor



import numpy as np
from math import floor, log10

# TODO openMDAO has generate_table() that can eventually replace this

# TODO this might have other use cases, move to utils if so


def round_it(x, sig=None):
    # default sig figs to 2 decimal places out
    if not sig:
        sig = len(str(round(x)))+2
    if x != 0:
        return round(x, sig-int(floor(log10(abs(x))))-1)
    else:
        return 0


def write_markdown_variable_table(open_file, problem, outputs, metadata):
    """
    Writes a table of the provided variable names in outputs. Converts units to defaults
    from metadata if avaliable.
    """
    open_file.write('\n| Variable Name | Value | Units |\n')
    open_file.write('| :- | :- | :- |\n')
    for var_name in outputs:
        # get default units from metadata
        try:
            units = metadata[var_name]['units']
        except KeyError:
            units = None
        # get value from problem
        try:
            if units:
                val = problem.get_val(var_name, units)
            else:
                # TODO find units for variable in problem?
                val = problem.get_val(var_name)
                units = 'unknown'
        # variable not in problem, get from aviary_inputs instead
        except KeyError:
            try:
                if units:
                    val = problem.aviary_inputs.get_val(var_name, units)
                else:
                    val, units = problem.aviary_inputs.get_item(var_name)
            except KeyError:
                val = 'Not Found in Model'
                units = None
        # handle rounding + formatting
        if isinstance(val, (np.ndarray, list, tuple)):
            val = [round_it(item) for item in val]
            # if an interable with a length of 1, remove bracket/paretheses, etc.
            if len(val) == 1:
                val = val[0]
        else:
            round_it(val)
        if not units:
            units = 'unknown'
        summary_line = f'| {var_name} | {val} |'
        if units != 'unitless':
            summary_line = summary_line + f' {units}'
        summary_line = summary_line + ' |\n'
        open_file.write(summary_line)



import unittest
import os

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.geometry.gasp_based.size_group import SizeGroup
from aviary.variable_info.options import get_option_defaults
from aviary.variable_info.variables import Aircraft, Mission

# this is the GASP test case, input and output values based on large single aisle 1 v3 without bug fix


class SizeGroupTestCase1(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Electrical.HAS_HYBRID_SYSTEM,
                        val=False, units='unitless')
        options.set_val(Aircraft.CrewPayload.NUM_PASSENGERS, val=180, units='unitless')
        options.set_val(Aircraft.Fuselage.AISLE_WIDTH, 24, units="inch")
        options.set_val(Aircraft.Fuselage.NUM_AISLES, 1)
        options.set_val(Aircraft.Fuselage.NUM_SEATS_ABREAST, 6)
        options.set_val(Aircraft.Fuselage.SEAT_PITCH, 29, units="inch")
        options.set_val(Aircraft.Fuselage.SEAT_WIDTH, 20.2, units="inch")

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "size",
            SizeGroup(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, val=10.13, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.SWEEP, val=25, units="deg")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.LOADING, val=128, units="lbf/ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.ASPECT_RATIO, val=1.67, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.TAPER_RATIO, val=0.352, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.TAPER_RATIO, val=0.801, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALE_FACTOR, val=1.028233
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, 0.6, units="unitless"
        )

        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.VOLUME_COEFFICIENT, val=1.189, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.VOLUME_COEFFICIENT, 0.145, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.DELTA_DIAMETER, 4.5, units="ft"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH, 9.5, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.NOSE_FINENESS, 1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.TAIL_FINENESS, 3, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.WETTED_AREA_FACTOR, 4000, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_TIP, 0.12, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MOMENT_RATIO, val=0.2307, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MOMENT_RATIO, 2.362, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.ASPECT_RATIO, val=4.75, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Engine.REFERENCE_DIAMETER, 5.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.CORE_DIAMETER_RATIO, 1.25, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.FINENESS, 2, units="unitless")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(self.prob[Aircraft.Fuselage.AVG_DIAMETER], 157.2, tol)
        assert_near_equal(self.prob["fuselage.cabin_height"], 13.1, tol)
        assert_near_equal(self.prob["fuselage.cabin_len"], 72.1, tol)
        assert_near_equal(self.prob["fuselage.nose_height"], 8.6, tol)

        # note: this is the actual GASP value, but for version 3.5. Version 3 has 129.4
        assert_near_equal(
            self.prob[Aircraft.Fuselage.LENGTH], 129.5, tol
        )
        assert_near_equal(self.prob[Aircraft.Fuselage.WETTED_AREA], 4000, tol)
        # note: this is the actual GASP value, but for version 3.5. Version 3 has 129.4
        assert_near_equal(
            self.prob[Aircraft.TailBoom.LENGTH], 129.5, tol
        )

        assert_near_equal(self.prob[Aircraft.Wing.AREA], 1370.3, tol)
        assert_near_equal(self.prob[Aircraft.Wing.SPAN], 117.8, tol)

        assert_near_equal(self.prob[Aircraft.Wing.CENTER_CHORD], 17.49, tol)
        assert_near_equal(self.prob[Aircraft.Wing.AVERAGE_CHORD], 12.615, tol)
        assert_near_equal(self.prob[Aircraft.Wing.ROOT_CHORD], 16.41, tol)
        assert_near_equal(
            self.prob[Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED], 0.1397, tol
        )  # not exact GASP value, likely due to rounding error
        assert_near_equal(self.prob[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX], 1114, tol)

        assert_near_equal(self.prob[Aircraft.HorizontalTail.AREA], 375.9, tol)
        assert_near_equal(self.prob[Aircraft.HorizontalTail.SPAN], 42.25, tol)
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.ROOT_CHORD], 13.16130387591471, tol
        )
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.AVERAGE_CHORD], 9.57573, tol
        )
        assert_near_equal(self.prob[Aircraft.HorizontalTail.MOMENT_ARM], 54.7, tol)

        assert_near_equal(self.prob[Aircraft.VerticalTail.AREA], 469.3, tol)
        assert_near_equal(self.prob[Aircraft.VerticalTail.SPAN], 28, tol)
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.ROOT_CHORD], 18.61267549773935, tol
        )
        assert_near_equal(self.prob[Aircraft.VerticalTail.AVERAGE_CHORD], 16.83022, tol)
        assert_near_equal(self.prob[Aircraft.VerticalTail.MOMENT_ARM], 49.9, tol)

        assert_near_equal(self.prob[Aircraft.Nacelle.AVG_DIAMETER], 7.35, tol)
        assert_near_equal(self.prob[Aircraft.Nacelle.AVG_LENGTH], 14.7, tol)
        assert_near_equal(self.prob[Aircraft.Nacelle.SURFACE_AREA], 339.58, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=2e-12, rtol=1e-12)


class SizeGroupTestCase2(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Fuselage.PROVIDE_SURFACE_AREA,
                        val=False, units='unitless')
        options.set_val(Aircraft.Wing.HAS_FOLD, val=True, units='unitless')
        options.set_val(Aircraft.Wing.HAS_STRUT, val=True, units='unitless')
        options.set_val(Aircraft.Electrical.HAS_HYBRID_SYSTEM,
                        val=False, units='unitless')
        options.set_val(Aircraft.CrewPayload.NUM_PASSENGERS, val=180, units='unitless')
        options.set_val(Aircraft.Wing.CHOOSE_FOLD_LOCATION, val=False, units='unitless')
        options.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')
        options.set_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')
        options.set_val(Aircraft.Fuselage.AISLE_WIDTH, 24, units="inch")
        options.set_val(Aircraft.Fuselage.NUM_AISLES, 1)
        options.set_val(Aircraft.Fuselage.NUM_SEATS_ABREAST, 6)
        options.set_val(Aircraft.Fuselage.SEAT_PITCH, 29, units="inch")
        options.set_val(Aircraft.Fuselage.SEAT_WIDTH, 20.2, units="inch")

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "size",
            SizeGroup(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, val=10.13, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.SWEEP, val=25, units="deg")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.LOADING, val=128, units="lbf/ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Strut.ATTACHMENT_LOCATION, val=12, units='ft'
        )
        self.prob.model.set_input_defaults(
            Aircraft.Strut.AREA_RATIO, val=0.021893, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.ASPECT_RATIO, val=1.67, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.TAPER_RATIO, val=0.801, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALE_FACTOR, val=1.028233
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, 0.6, units="unitless"
        )

        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.VOLUME_COEFFICIENT, val=1.189, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.VOLUME_COEFFICIENT, 0.145, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.DELTA_DIAMETER, 4.5, units="ft"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH, 9.5, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.NOSE_FINENESS, 1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.TAIL_FINENESS, 3, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.WETTED_AREA_FACTOR, 4000, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_TIP, 0.12, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MOMENT_RATIO, val=0.2307, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MOMENT_RATIO, 2.362, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.ASPECT_RATIO, val=4.75, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Engine.REFERENCE_DIAMETER, 5.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.CORE_DIAMETER_RATIO, 1.25, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.FINENESS, 2, units="unitless")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Fuselage.AVG_DIAMETER], 157.2, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["fuselage.cabin_height"], 13.1, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["fuselage.cabin_len"], 72.09722222, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["fuselage.nose_height"], 8.6, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Fuselage.LENGTH], 129.5, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Fuselage.WETTED_AREA], 18558260.55555555, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.TailBoom.LENGTH], 129.5, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Wing.AREA], 1370.3125, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.SPAN], 117.81878299, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Wing.CENTER_CHORD], 17.48974356, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.AVERAGE_CHORD], 12.61453233, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.ROOT_CHORD], 16.40711451, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED], 0.13965584, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob["wing.fold.nonfolded_taper_ratio"], 0.93175961, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.FOLDING_AREA], 1167.5966191, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["wing.fold.nonfolded_wing_area"], 202.7158809, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["wing.fold.tc_ratio_mean_folded"], 0.14847223, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["wing.fold.nonfolded_AR"], 0.71035382, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX], 208.08091725, tol
        )  # not actual GASP value

        assert_near_equal(self.prob["wing.strut_y"], 6, tol)  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Strut.LENGTH], 13.11154072, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Strut.CHORD], 1.14403031, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.AREA], 375.87987047, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.SPAN], 42.25434161, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.ROOT_CHORD], 13.15924684, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.AVERAGE_CHORD], 9.57681709, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.MOMENT_ARM], 54.67937726, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.VerticalTail.AREA], 469.31832812, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.SPAN], 27.99574268, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.ROOT_CHORD], 18.61623295, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.AVERAGE_CHORD], 16.83214111, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.MOMENT_ARM], 49.88094115, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Nacelle.AVG_DIAMETER], 7.35163168, tol
        )  # may not be actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Nacelle.AVG_LENGTH], 14.70326336, tol
        )  # may not be actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Nacelle.SURFACE_AREA], 339.58410134, tol
        )  # may not be actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=3e-10, rtol=1e-12)


class SizeGroupTestCase3(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Fuselage.PROVIDE_SURFACE_AREA,
                        val=False, units='unitless')
        options.set_val(Aircraft.Wing.HAS_FOLD, val=True, units='unitless')
        options.set_val(Aircraft.Design.COMPUTE_HTAIL_VOLUME_COEFF,
                        val=True, units='unitless')
        options.set_val(Aircraft.Design.COMPUTE_VTAIL_VOLUME_COEFF,
                        val=True, units='unitless')
        options.set_val(Aircraft.CrewPayload.NUM_PASSENGERS, val=180, units='unitless')
        options.set_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')
        options.set_val(Aircraft.Fuselage.AISLE_WIDTH, 24, units="inch")
        options.set_val(Aircraft.Fuselage.NUM_AISLES, 1)
        options.set_val(Aircraft.Fuselage.NUM_SEATS_ABREAST, 1)
        options.set_val(Aircraft.Fuselage.SEAT_PITCH, 29, units="inch")
        options.set_val(Aircraft.Fuselage.SEAT_WIDTH, 20.2, units="inch")

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "size",
            SizeGroup(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, val=10.13, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.SWEEP, val=25, units="deg")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.LOADING, val=128, units="lbf/ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FOLDED_SPAN, val=25, units="ft"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION, val=0, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.ASPECT_RATIO, val=1.67, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.TAPER_RATIO, val=0.352, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.TAPER_RATIO, val=0.801, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALE_FACTOR, val=1.028233
        )
        self.prob.model.set_input_defaults(
            Aircraft.Engine.WING_LOCATIONS, val=0.35, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, 0.6, units="unitless"
        )

        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.VOLUME_COEFFICIENT, val=1.189, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.VOLUME_COEFFICIENT, 0.145, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.DELTA_DIAMETER, 4.5, units="ft"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH, 9.5, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.NOSE_FINENESS, 1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.TAIL_FINENESS, 3, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.WETTED_AREA_FACTOR, 4000, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_TIP, 0.12, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MOMENT_RATIO, val=0.2307, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MOMENT_RATIO, 2.362, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.ASPECT_RATIO, val=4.75, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Engine.REFERENCE_DIAMETER, 5.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.CORE_DIAMETER_RATIO, 1.25, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.FINENESS, 2, units="unitless")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Fuselage.AVG_DIAMETER], 56.2, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["fuselage.cabin_height"], 9.18333, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["fuselage.cabin_len"], 435, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["fuselage.nose_height"], 4.68333, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Fuselage.LENGTH], 476.7333, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Fuselage.WETTED_AREA], 53601769, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.TailBoom.LENGTH], 476.7333, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Wing.AREA], 1370.3125, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.SPAN], 117.81878299, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Wing.CENTER_CHORD], 17.48974356, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.AVERAGE_CHORD], 12.61453233, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.ROOT_CHORD], 16.988, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED], 0.14151, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob["wing.fold.nonfolded_taper_ratio"], 0.85783252, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.FOLDING_AREA], 964.14982163, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["wing.fold.nonfolded_wing_area"], 406.16267837, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["wing.fold.tc_ratio_mean_folded"], 0.14681715, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["wing.fold.nonfolded_AR"], 1.5387923, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX], 406.53567274, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.AREA], 298.484, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.SPAN], 37.654, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.ROOT_CHORD], 11.7265, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.AVERAGE_CHORD], 8.5341, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.MOMENT_ARM], 54.67937726, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.VerticalTail.AREA], 297.003, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.SPAN], 22.2709, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.ROOT_CHORD], 14.8094, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.AVERAGE_CHORD], 13.3902, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.MOMENT_ARM], 49.88094115, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Nacelle.AVG_DIAMETER], 7.35163168, tol
        )  # may not be actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Nacelle.AVG_LENGTH], 14.70326336, tol
        )  # may not be actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Nacelle.SURFACE_AREA], 339.58410134, tol
        )  # may not be actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Electrical.HYBRID_CABLE_LENGTH], 50.6032, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-9, rtol=1e-12)


class SizeGroupTestCase4(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Design.COMPUTE_HTAIL_VOLUME_COEFF,
                        val=True, units='unitless')
        options.set_val(Aircraft.Design.COMPUTE_VTAIL_VOLUME_COEFF,
                        val=True, units='unitless')
        options.set_val(Aircraft.Wing.HAS_STRUT, val=True, units='unitless')
        options.set_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED,
                        val=False, units='unitless')
        options.set_val(Aircraft.Electrical.HAS_HYBRID_SYSTEM,
                        val=False, units='unitless')
        options.set_val(Aircraft.CrewPayload.NUM_PASSENGERS, val=180, units='unitless')
        options.set_val(Aircraft.Fuselage.AISLE_WIDTH, 24, units="inch")
        options.set_val(Aircraft.Fuselage.NUM_AISLES, 1)
        options.set_val(Aircraft.Fuselage.NUM_SEATS_ABREAST, 1)
        options.set_val(Aircraft.Fuselage.SEAT_PITCH, 29, units="inch")
        options.set_val(Aircraft.Fuselage.SEAT_WIDTH, 20.2, units="inch")

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "size",
            SizeGroup(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, val=10.13, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )
        self.prob.model.set_input_defaults(Aircraft.Wing.SWEEP, val=25, units="deg")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.LOADING, val=128, units="lbf/ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Strut.ATTACHMENT_LOCATION_DIMENSIONLESS, val=0, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Strut.AREA_RATIO, val=0.021893, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION, val=0, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.ASPECT_RATIO, val=1.67, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.TAPER_RATIO, val=0.352, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.TAPER_RATIO, val=0.801, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALE_FACTOR, val=1.028233
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuel.WING_FUEL_FRACTION, 0.6, units="unitless"
        )

        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.VOLUME_COEFFICIENT, val=1.189, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.VOLUME_COEFFICIENT, 0.145, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.DELTA_DIAMETER, 4.5, units="ft"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.PILOT_COMPARTMENT_LENGTH, 9.5, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.NOSE_FINENESS, 1, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.TAIL_FINENESS, 3, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.WETTED_AREA_FACTOR, 4000, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_TIP, 0.12, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MOMENT_RATIO, val=0.2307, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MOMENT_RATIO, 2.362, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.ASPECT_RATIO, val=4.75, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Engine.REFERENCE_DIAMETER, 5.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.CORE_DIAMETER_RATIO, 1.25, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.FINENESS, 2, units="unitless")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Fuselage.AVG_DIAMETER], 56.2, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["fuselage.cabin_height"], 9.18333, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["fuselage.cabin_len"], 435, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["fuselage.nose_height"], 4.68333, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Fuselage.LENGTH], 476.7333, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Fuselage.WETTED_AREA], 4000, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.TailBoom.LENGTH], 476.7333, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Wing.AREA], 1370.3125, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.SPAN], 117.81878299, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Wing.CENTER_CHORD], 17.48974356, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.AVERAGE_CHORD], 12.61453233, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.ROOT_CHORD], 16.988, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED], 0.14151, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob["wing.strut.strut_y"], 0, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Strut.LENGTH], 5.2361, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Strut.CHORD], 2.8647, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.AREA], 298.484, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.SPAN], 37.654, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.ROOT_CHORD], 11.7265, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.AVERAGE_CHORD], 8.5341, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.MOMENT_ARM], 54.67937726, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.VerticalTail.AREA], 297.003, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.SPAN], 22.2709, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.ROOT_CHORD], 14.8094, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.AVERAGE_CHORD], 13.3902, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.MOMENT_ARM], 49.88094115, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Nacelle.AVG_DIAMETER], 7.35163168, tol
        )  # may not be actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Nacelle.AVG_LENGTH], 14.70326336, tol
        )  # may not be actual GASP value
        assert_near_equal(
            self.prob[Aircraft.Nacelle.SURFACE_AREA], 339.58410134, tol
        )  # may not be actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=2e-12, rtol=1e-12)


if __name__ == "__main__":
    unittest.main()
